#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools mkdepthcharge program
# Copyright (C) 2020-2023 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

import argparse
import logging
import os
import platform
import struct
import subprocess
import sys
import tempfile

from mmap import mmap
from pathlib import Path

from depthcharge_tools import __version__
from depthcharge_tools.utils.argparse import (
    Command,
    Argument,
    Group,
)
from depthcharge_tools.utils.pathlib import (
    copy,
    decompress,
)
from depthcharge_tools.utils.platform import (
    Architecture,
    vboot_keys,
)
from depthcharge_tools.utils.string import (
    parse_bytesize,
)
from depthcharge_tools.utils.subprocess import (
    mkimage,
    vbutil_kernel,
    lz4,
    lzma,
    fdtget,
    fdtput,
)


class mkdepthcharge(
    Command,
    prog="mkdepthcharge",
    usage="%(prog)s [options] -o FILE [--] [VMLINUZ] [INITRAMFS ...] [DTB ...]",
    add_help=False,
):
    """Build boot images for the ChromeOS bootloader."""

    _logger = logging.getLogger(__name__)

    @property
    def logger(self):
        # Set verbosity before logging messages
        self.verbosity
        return self._logger

    # Inputs can have the same name and cause collisions in tmpdir.
    def _tempfile(self, name):
        f = self.tmpdir / name
        if not f.exists():
            return f

        for i in range(9999):
            f = f.with_name("{}-idx{:04}".format(name, i))
            if not f.exists():
                return f

        raise FileExistsError(self.tmpdir / name)

    # Debian packs the arm64 kernel uncompressed, but the bindeb-pkg
    # kernel target packs it as gzip. So we'll try to decompress inputs.
    def _decompress(self, f):
        self.logger.info(
            "Trying to decompress file '{}'."
            .format(f)
        )

        decomp = decompress(f, self._tempfile(f.name))
        if decomp is not None:
            self.logger.info(
                "Decompressed input '{}' as '{}'."
                .format(f, decomp)
            )

        return decomp or f

    # Copy inputs to tmpdir because mkimage wants modifiable files.
    def _copy(self, f):
        if self.tmpdir not in f.parents:
            f = copy(f, self._tempfile(f.name))
        f.chmod(0o755)

        return f

    @Group
    def input_files(self):
        """Input files"""

    @input_files.add
    @Argument(help=argparse.SUPPRESS)
    def files(self, *files):
        """Input files (vmlinuz, initramfs, dtbs)"""
        vmlinuz = []
        initramfs = []
        dtbs = []

        files = [Path(f).resolve() for f in files]

        for f in files:
            # Decompress files to run detection on content.
            decomp = self._decompress(f)
            with decomp.open("rb") as f_:
                head = f_.read(4096)

            # Portable Executable and ELF files
            if head.startswith(b"MZ") or head.startswith(b"ELF"):
                self.logger.info(
                    "File '{}' identified as a vmlinuz."
                    .format(f)
                )
                vmlinuz.append(decomp)

            # Cpio files
            elif (
                head.startswith(b"070701")
                or head.startswith(b"070702")
                or head.startswith(b"070707")
            ):
                self.logger.info(
                    "File '{}' identified as an initramfs."
                    .format(f)
                )
                # Keep initramfs compressed
                initramfs.append(f)

                # Avoid name collision when copying initramfs to tmpdir
                if decomp != f:
                    decomp.unlink()

            # Device-tree blobs
            elif head.startswith(b"\xd0\x0d\xfe\xed"):
                self.logger.info(
                    "File '{}' identified as a device-tree blob."
                    .format(f)
                )
                dtbs.append(decomp)

            # Failed to detect, assume in the order in usage string
            elif len(vmlinuz) == 0:
                self.logger.info(
                    "Assuming file '{}' is a vmlinuz."
                    .format(f)
                )
                vmlinuz.append(decomp)

            elif len(initramfs) == 0:
                self.logger.info(
                    "Assuming file '{}' is an initramfs."
                    .format(f)
                )
                # Keep initramfs compressed
                initramfs.append(f)

            else:
                self.logger.info(
                    "Assuming file '{}' is a device-tree blob."
                    .format(f)
                )
                dtbs.append(decomp)

        return {
            "vmlinuz": vmlinuz,
            "initramfs": initramfs,
            "dtbs": dtbs,
        }

    @input_files.add
    @Argument("-d", "--vmlinuz", nargs=1)
    def vmlinuz(self, vmlinuz=None):
        """Kernel executable"""
        files = self.files["vmlinuz"]

        if vmlinuz is not None:
            vmlinuz = Path(vmlinuz).resolve()
            self.logger.info(
                "Using file '{}' as a vmlinuz."
                .format(vmlinuz)
            )
            files = [self._decompress(vmlinuz), *files]

        if len(files) == 0:
            raise ValueError(
                "vmlinuz argument is required."
            )

        elif len(files) > 1:
            raise ValueError(
                "Can't build with multiple kernels"
            )

        vmlinuz = self._copy(files[0])

        return vmlinuz

    @input_files.add
    @Argument("-i", "--initramfs", metavar="INITRAMFS", nargs="+")
    def initramfs(self, *files):
        """Ramdisk images"""
        files = [
            *(Path(f).resolve() for f in files if f is not None),
            *self.files["initramfs"],
        ]

        for file in files:
            self.logger.info(
                "Using file '{}' as an initramfs."
                .format(file)
            )

        if len(files) > 1:
            self.logger.info(
                "Concatenating initramfs files as a single initramfs."
            )
            initramfs = self._tempfile("merged-initramfs.img")
            with initramfs.open('xb') as merge:
                for file in files:
                    merge.write(file.read_bytes())

        elif len(files) == 1:
            initramfs = self._copy(files[0])

        elif not files:
            initramfs = None

        return initramfs

    @input_files.add
    @Argument("-b", "--dtbs", metavar="DTB", nargs="+")
    def dtbs(self, *dtbs):
        """Device-tree binary file"""
        files = self.files["dtbs"]

        dtbs = [Path(dtb).resolve() for dtb in dtbs]
        for dtb in dtbs:
            self.logger.info(
                "Using file '{}' as a device-tree blob."
                .format(dtb)
            )

        dtbs = [self._decompress(dtb) for dtb in dtbs]
        dtbs = [self._copy(dtb) for dtb in (*dtbs, *files)]

        return dtbs

    @Group
    def options(self):
        """Options"""
        # Check incompatible combinations
        if self.image_format == "zimage":
            if self.compress not in (None, "none"):
                raise ValueError(
                    "Compress argument not supported with zimage format."
                )
            if self.name is not None:
                raise ValueError(
                    "Name argument not supported with zimage format."
                )
            if self.dtbs:
                raise ValueError(
                    "Device tree files not supported with zimage format."
                )

    @options.add
    @Argument("-h", "--help", action="help")
    def print_help(self):
        """Show this help message."""
        # type(self).parser.print_help()

    @options.add
    @Argument(
        "-V", "--version",
        action="version",
        version="depthcharge-tools %(prog)s {}".format(__version__),
    )
    def version(self):
        """Print program version."""
        return type(self).version.version % {"prog": type(self).prog}

    @options.add
    @Argument("-v", "--verbose", count=True)
    def verbosity(self, verbosity=0):
        """Print more detailed output."""
        level = logging.WARNING - int(verbosity) * 10
        self._logger.setLevel(level)
        return verbosity

    @options.add
    @Argument("-o", "--output", required=True)
    def output(self, file_):
        """Write resulting image to FILE."""

        # Output path is obviously required
        if file_ is None:
            raise ValueError(
                "Output argument is required."
            )

        return Path(file_).resolve()

    @options.add
    @Argument("--tmpdir", nargs=1)
    def tmpdir(self, dir_=None):
        """Directory to keep temporary files."""
        if dir_ is None:
            dir_ = tempfile.TemporaryDirectory(
                prefix="mkdepthcharge-",
            )
            dir_ = self.exitstack.enter_context(dir_)

        dir_ = Path(dir_)
        os.makedirs(dir_, exist_ok=True)

        self.logger.debug("Working in temp dir '{}'.".format(dir_))

        return dir_

    @options.add
    @Argument("-A", "--arch", nargs=1)
    def arch(self, arch=None):
        """Architecture to build for."""

        # We should be able to make an image for other architectures, but
        # the default should be whatever board the kernel is for.
        if arch is None:
            with self.vmlinuz.open("rb") as f:
                head = f.read(4096)

            if head[0x202:0x206] == b"HdrS":
                arch = Architecture("x86")
            elif head[0x38:0x3c] == b"ARM\x64":
                arch = Architecture("arm64")
            elif head[0x34:0x38] == b"\x45\x45\x45\x45":
                arch = Architecture("arm")

            self.logger.info(
                "Assuming CPU architecture '{}' from vmlinuz file."
                .format(arch)
            )

        elif arch not in Architecture.all:
            raise ValueError(
                "Can't build images for unknown architecture '{}'"
                .format(arch)
            )

        return Architecture(arch)

    @options.add
    @Argument("--format", nargs=1)
    def image_format(self, format_=None):
        """Kernel image format to use."""

        # Default to architecture-specific formats.
        if format_ is None:
            if self.arch in Architecture.arm:
                format_ = "fit"
            elif self.arch in Architecture.x86:
                format_ = "zimage"
            self.logger.info("Assuming image format '{}'.".format(format_))

        if format_ not in ("fit", "zimage"):
            raise ValueError(
                "Can't build images for unknown image format '{}'"
                .format(format_)
            )

        return format_

    @options.add
    @Argument("--kernel-start", nargs=1)
    def kernel_start(self, addr=None):
        """Start of depthcharge kernel buffer in memory."""
        if addr is not None:
            return parse_bytesize(addr)

        if self.arch in Architecture.x86:
            return 0x100000

    @options.add
    @Argument(
        "--no-pad-vmlinuz", pad=False,
        help=argparse.SUPPRESS,
    )
    @Argument(
        "--pad-vmlinuz", pad=True,
        help="Pad vmlinuz for safe decompression.",
    )
    def pad_vmlinuz(self, pad=None):
        """Pad vmlinuz for safe decompression."""
        if pad is None:
            pad = (
                self.image_format == "fit"
                and self.patch_dtbs
            )

        return bool(pad)

    @Group
    def fit_options(self):
        """FIT image options"""

    @fit_options.add
    @Argument("-C", "--compress", nargs=1)
    def compress(self, type_=None):
        """Compress vmlinuz file before packing."""

        # We need to pass "-C none" to mkimage or it assumes gzip.
        if type_ is None and self.image_format == "fit":
            type_ = "none"

        if type_ not in (None, "none", "lz4", "lzma"):
            raise ValueError(
                "Compression type '{}' is not supported."
                .format(type_)
            )

        return type_

    @fit_options.add
    @Argument("-n", "--name", nargs=1)
    def name(self, desc=None):
        """Description of vmlinuz to put in the FIT."""

        # If we don't pass "-n <name>" to mkimage, the kernel image
        # description is left blank. Other images get "unavailable"
        # as their description, so it looks better if we match that.
        if desc is None and self.image_format == "fit":
            desc = "unavailable"

        return desc

    @fit_options.add
    @Argument("--ramdisk-load-address", nargs=1)
    def ramdisk_load_address(self, addr=None):
        """Add load address to FIT ramdisk image section."""
        if addr is not None:
            return parse_bytesize(addr)

        return None

    @fit_options.add
    @Argument(
        "--no-patch-dtbs", patch_dtbs=False,
        help=argparse.SUPPRESS,
    )
    @Argument(
        "--patch-dtbs", patch_dtbs=True,
        help="Add linux,initrd properties to device-tree binary files.",
    )
    def patch_dtbs(self, patch_dtbs=False):
        """Add linux,initrd properties to device-tree binary files."""
        if (
            patch_dtbs
            and self.kernel_start is None
            and self.ramdisk_load_address is None
        ):
            raise ValueError(
                "The kernel buffer start address or a ramdisk load address "
                "is required to patch DTB files for initramfs support."
            )

        return bool(patch_dtbs)

    @Group
    def zimage_options(self):
        """zImage format options"""

    @zimage_options.add
    @Argument(
        "--no-set-init-size", init_size=False,
        help="Don't set init_size boot param.",
    )
    @Argument(
        "--set-init-size", init_size=True,
        help=argparse.SUPPRESS,
    )
    def set_init_size(self, init_size=None):
        """Set init_size boot param for safe decompression."""
        if init_size is None:
            return (
                self.image_format == "zimage"
                and self.initramfs is not None
            )

        return bool(init_size)

    @Group
    def vboot_options(self):
        """Depthcharge image options"""

        keydirs = []
        if self.keydir is not None:
            keydirs += [self.keydir]

        # If any of the arguments are given, search nearby for others
        if self.keyblock is not None:
            keydirs += [self.keyblock.parent]
        if self.signprivate is not None:
            keydirs += [self.signprivate.parent]
        if self.signpubkey is not None:
            keydirs += [self.signpubkey.parent]

        if None in (self.keyblock, self.signprivate, self.signpubkey):
            for d in sorted(set(keydirs), key=keydirs.index):
                self.logger.info(
                    "Searching '{}' for vboot keys."
                    .format(d)
                )

            # Defaults to distro-specific paths for necessary files.
            keydir, keyblock, signprivate, signpubkey = vboot_keys(*keydirs)

            if keydir:
                self.logger.info(
                    "Defaulting to keys from '{}' for missing arguments."
                    .format(keydir)
                )

            if self.keyblock is None:
                self.keyblock = keyblock
            if self.signprivate is None:
                self.signprivate = signprivate
            if self.signpubkey is None:
                self.signpubkey = signpubkey

        # We might still not have the vboot keys after all that.
        if self.keyblock is None:
            raise ValueError(
                "Couldn't find a usable keyblock file."
            )
        elif not self.keyblock.is_file():
            raise ValueError(
                "Keyblock file '{}' does not exist."
                .format(self.keyblock)
            )
        else:
            self.logger.info(
                "Using keyblock file '{}'."
                .format(self.keyblock)
            )

        if self.signprivate is None:
            raise ValueError(
                "Couldn't find a usable signprivate file."
            )
        elif not self.signprivate.is_file():
            raise ValueError(
                "Signprivate file '{}' does not exist."
                .format(self.signprivate)
            )
        else:
            self.logger.info(
                "Using signprivate file '{}'."
                .format(self.signprivate)
            )

        if self.signpubkey is None:
            self.logger.warning(
                "Couldn't find a usable signpubkey file."
            )
        elif not self.signpubkey.is_file():
            self.logger.warning(
                "Signpubkey file '{}' does not exist."
                .format(self.keyblock)
            )
            self.signpubkey = None
        else:
            self.logger.info(
                "Using signpubkey file '{}'."
                .format(self.signpubkey)
            )

    @vboot_options.add
    @Argument("-c", "--cmdline", append=True, nargs="+")
    def cmdline(self, *cmd):
        """Command-line parameters for the kernel."""

        # If the cmdline is empty vbutil_kernel returns an error. We can use
        # "--" instead of putting a newline or a space into the cmdline.
        if len(cmd) == 0:
            cmdline = "--"
        elif len(cmd) == 1 and isinstance(cmd[0], str):
            cmdline = cmd[0]
        elif isinstance(cmd, (list, tuple)):
            cmdline = " ".join(cmd)

        # The firmware replaces any '%U' in the kernel cmdline with the
        # PARTUUID of the partition it booted from. Chrome OS uses
        # kern_guid=%U in their cmdline and it's useful information, so
        # prepend it to cmdline.
        if (self.kern_guid is None) or self.kern_guid:
            cmdline = " ".join(("kern_guid=%U", cmdline))

        return cmdline

    @vboot_options.add
    @Argument(
        "--no-kern-guid", kern_guid=False,
        help="Don't prepend kern_guid=%%U to the cmdline."
    )
    @Argument(
        "--kern-guid", kern_guid=True,
        help=argparse.SUPPRESS,
    )
    def kern_guid(self, kern_guid=True):
        """Prepend kern_guid=%%U to the cmdline."""
        return kern_guid

    @vboot_options.add
    @Argument("--bootloader", nargs=1)
    def bootloader(self, file_=None):
        """Bootloader stub binary to use."""
        if file_ is not None:
            file_ = Path(file_).resolve()

        if (
            self.image_format == "zimage"
            and self.initramfs is not None
            and file_ is not None
        ):
            raise ValueError(
                "Can't build images with both initramfs and "
                "bootloader stub for zimage format."
            )

        return file_

    @vboot_options.add
    @Argument("--keydir")
    def keydir(self, dir_):
        """Directory containing vboot keys to use."""
        if dir_ is not None:
            dir_ = Path(dir_).resolve()

        return dir_

    @vboot_options.add
    @Argument("--keyblock")
    def keyblock(self, file_):
        """The key block file (.keyblock)."""
        if file_ is not None:
            file_ = Path(file_).resolve()

        return file_

    @vboot_options.add
    @Argument("--signprivate")
    def signprivate(self, file_):
        """Private key (.vbprivk) to sign the image."""
        if file_ is not None:
            file_ = Path(file_).resolve()

        return file_

    @vboot_options.add
    @Argument("--signpubkey")
    def signpubkey(self, file_):
        """Public key (.vbpubk) to verify the image."""
        if file_ is not None:
            file_ = Path(file_).resolve()

        return file_

    def __call__(self):
        vmlinuz = self.vmlinuz
        initramfs = self.initramfs
        bootloader = self.bootloader
        dtbs = self.dtbs
        tmpdir = self.tmpdir

        if bootloader is not None:
            bootloader = copy(bootloader, tmpdir)

        # Depthcharge on arm64 with FIT supports these two compressions.
        if self.compress == "lz4":
            self.logger.info("Compressing kernel with lz4.")
            vmlinuz = lz4.compress(vmlinuz, self._tempfile("vmlinuz.lz4"))
        elif self.compress == "lzma":
            self.logger.info("Compressing kernel with lzma.")
            vmlinuz = lzma.compress(vmlinuz, self._tempfile("vmlinuz.lzma"))
        elif self.compress not in (None, "none"):
            fmt = "Compression type '{}' is not supported."
            msg = fmt.format(compress)
            raise ValueError(msg)

        # vbutil_kernel --config argument wants cmdline as a file.
        cmdline_file = self._tempfile("kernel.args")
        cmdline_file.write_text(self.cmdline)

        # vbutil_kernel --bootloader argument is mandatory, but it's
        # unused in depthcharge except as a multiboot ramdisk. Prepare
        # this empty file as its replacement where necessary.
        empty = self._tempfile("empty.bin")
        empty.write_bytes(bytes(512))

        # The kernel decompression overwrites parts of the buffer we
        # control while decompressing itself. We need to make sure we
        # don't place initramfs in that range. For that, we need to know
        # how offsets in file correspond to addresses in memory.

        def addr_to_offs(addr, load_addr=self.kernel_start):
            return addr - load_addr + 0x10000

        def offs_to_addr(offs, load_addr=self.kernel_start):
            return offs + load_addr - 0x10000

        def align_up(size, align=0x1000):
            return ((size + align - 1) // align) * align

        # Size for a small padding, sometimes necessary in some
        # places for unknown reasons, added and set empirically.
        small_pad = 0x40000

        if self.image_format == "fit":
            fit_image = self._tempfile("depthcharge.fit")

            initramfs_args = []
            if initramfs is not None:
                initramfs_args += ["-i", initramfs]

            dtb_args = []
            for dtb in dtbs:
                dtb_args += ["-b", dtb]

            # The subimage nodes can be <type>@1 or <type>-1.
            def subimage_by_type(fit_image, subimage_type):
                for subimage in fdtget.subnodes(fit_image, "/images"):
                    node = "/images/{}".format(subimage)
                    try:
                        if fdtget.get(fit_image, node, "type") == subimage_type:
                            return node
                    except:
                        continue

            # On later 32-bit ARM Chromebooks, the KERNEL_START address
            # can be very close to the where kernel decompresses itself
            # that the process overwrites the initramfs. The device-tree
            # is luckily copied away before then. We need to add some
            # vmlinuz padding to prevent this.
            if initramfs is not None and self.pad_vmlinuz:

                # We need the decompressed kernel size, not easy to get.
                # Try to find the compressed vmlinux inside vmlinuz,
                # then try to decompress it.
                data = vmlinuz.read_bytes()
                vmlinuz_size = len(data)
                decomp_size = -1

                for fmt, magic in {
                    "gzip":  b'\x1f\x8b\x08',
                    "xz":    b'\xfd7zXZ\x00',
                    "zstd":  b'(\xb5/\xfd',
                    "lzma":  b'\x5d\x00\x00\x00',
                    "lz4":   b'\02!L\x18',
                    "bzip2": b'BZh',
                    "lzop":  b'\x89\x4c\x5a',
                }.items():
                    offs = data.find(magic)
                    while 0 < offs < vmlinuz_size:
                        decomp = decompress(data[offs:], partial=True)
                        if decomp:
                            self.logger.info(
                                "Found {} at {:#x} in vmlinuz, with size {:#x}."
                                .format(fmt, offs, len(decomp))
                            )
                            decomp_size = max(decomp_size, len(decomp))
                        offs = data.find(magic, offs + 1)

                if decomp_size == -1:
                    raise ValueError(
                        "Couldn't find decompressed kernel inside vmlinuz."
                    )

                self.logger.info(
                    "Vmlinuz size is {:#x}, {:#x} decompressed."
                    .format(vmlinuz_size, decomp_size)
                )

                # Decompression starts at start of physical memory,
                # calculated per AUTO_ZRELADDR. But first kernel copies
                # itself after where the decompressed copy would end.
                decomp_addr = self.kernel_start & 0xf8000000
                safe_initrd_start = (
                    decomp_addr + decomp_size + vmlinuz_size + small_pad
                )
                initrd_start = (
                    self.kernel_start + vmlinuz_size
                    + sum(dtb.stat().st_size for dtb in self.dtbs)
                )

                if initrd_start < safe_initrd_start:
                    pad_to = align_up(
                        vmlinuz_size
                        + (safe_initrd_start - initrd_start)
                    )
                    self.logger.info(
                        "Padding vmlinuz to {:#x}."
                        .format(pad_to)
                    )
                    with vmlinuz.open("r+b") as f, mmap(f.fileno(), 0) as data:
                        data.resize(pad_to)

            # The later 32-bit ARM Chromebooks use Depthcharge, but
            # their stock versions don't have the code to support FIT
            # ramdisks. But since we know the fixed KERNEL_START we can
            # deduce where the initramfs will be, and inject its address
            # into the DTBs the way Linux expects bootloaders to do.
            if initramfs is not None and self.patch_dtbs:

                # We'll probably never need this, as only old U-Boot builds
                # need a ramdisk load address and those can handle the
                # initrd properties fine.
                if self.ramdisk_load_address:
                    initrd_start = self.ramdisk_load_address
                    initrd_end = initrd_start + initramfs.stat().st_size

                else:
                    # Allocate space for the properties we want to set,
                    # adding them later would shift things around.
                    self.logger.info("Preparing dtb files for initramfs support.")
                    for dtb in dtbs:
                        fdtput.put(dtb, "/chosen", "linux,initrd-start", 0)
                        fdtput.put(dtb, "/chosen", "linux,initrd-end", 0)

                    # Make a temporary image and search for the initramfs
                    # inside it, because I don't want to risk a wrong
                    # estimate and don't want to mess with pylibfdt.
                    self.logger.info("Packing files as temp FIT image:")
                    tmp_image = self._tempfile("depthcharge.fit.tmp")
                    proc = mkimage(
                        "-f", "auto",
                        "-A", self.arch.mkimage,
                        "-T", "kernel",
                        "-O", "linux",
                        "-C", self.compress,
                        "-n", self.name,
                        *initramfs_args,
                        *dtb_args,
                        "-d", vmlinuz,
                        tmp_image,
                    )
                    self.logger.info(proc.stdout)

                    # Mkimage breaks the config node key with -T kernel_noload.
                    # Apparently this shifts things around as well, so...
                    self.logger.info("Patching temp FIT for kernel_noload type.")
                    fdtput.put(
                        tmp_image, subimage_by_type(tmp_image, "kernel"),
                        "type", "kernel_noload",
                    )

                    with tmp_image.open("r+b") as f, mmap(f.fileno(), 0) as data:
                        initrd_offset = data.find(initramfs.read_bytes())
                        self.logger.info(
                            "Initramfs is at offset {:#x} in FIT image."
                            .format(initrd_offset)
                        )
                    initrd_start = initrd_offset + self.kernel_start
                    initrd_end = initrd_start + initramfs.stat().st_size

                self.logger.info(
                    "Initramfs should be at address {:#x} - {:#x} in memory."
                    .format(initrd_start, initrd_end)
                )

                self.logger.info("Patching dtb files for initramfs support.")
                for dtb in dtbs:
                    fdtput.put(dtb, "/chosen", "linux,initrd-start", initrd_start)
                    fdtput.put(dtb, "/chosen", "linux,initrd-end", initrd_end)

            self.logger.info("Packing files as FIT image:")
            proc = mkimage(
                "-f", "auto",
                "-A", self.arch.mkimage,
                "-T", "kernel",
                "-O", "linux",
                "-C", self.compress,
                "-n", self.name,
                *initramfs_args,
                *dtb_args,
                "-d", vmlinuz,
                fit_image,
            )
            self.logger.info(proc.stdout)

            # Earlier 32-bit ARM Chromebooks use U-Boot, which needs a
            # usable load address for the FIT ramdisk image section.
            if initramfs is not None and self.ramdisk_load_address:
                self.logger.info("Patching FIT for ramdisk load address.")
                fdtput.put(
                    fit_image, subimage_by_type(fit_image, "ramdisk"),
                    "load", self.ramdisk_load_address,
                )

            # Mkimage breaks the config node key with -T kernel_noload.
            self.logger.info("Patching FIT for kernel_noload type.")
            fdtput.put(
                fit_image, subimage_by_type(fit_image, "kernel"),
                "type", "kernel_noload",
            )

            if (
                initramfs is not None
                and self.patch_dtbs
                and self.ramdisk_load_address is None
            ):
                with fit_image.open("r+b") as f, mmap(f.fileno(), 0) as data:
                    if initrd_offset != data.find(initramfs.read_bytes()):
                        raise RuntimeError(
                            "Initramfs FIT offset changed after rebuild."
                        )

            self.logger.info("Packing files as depthcharge image.")
            proc = vbutil_kernel(
                "--version", "1",
                "--arch", self.arch.vboot,
                "--vmlinuz", fit_image,
                "--config", cmdline_file,
                "--bootloader", bootloader or empty,
                "--keyblock", self.keyblock,
                "--signprivate", self.signprivate,
                "--pack", self.output,
            )
            self.logger.info(proc.stdout)

        elif self.image_format == "zimage" and initramfs is None:
            self.logger.info("Packing files as depthcharge image.")
            proc = vbutil_kernel(
                "--version", "1",
                "--arch", self.arch.vboot,
                "--vmlinuz", vmlinuz,
                "--config", cmdline_file,
                "--bootloader", bootloader or empty,
                "--keyblock", self.keyblock,
                "--signprivate", self.signprivate,
                "--pack", self.output,
            )
            self.logger.info(proc.stdout)

        elif self.image_format == "zimage":
            # bzImage header has the address the kernel will decompress
            # to, and the amount of memory it needs there to work.
            # See Documentation/x86/boot.rst in Linux tree for offsets.
            with vmlinuz.open("r+b") as f, mmap(f.fileno(), 0) as data:
                if data[0x202:0x206] != b"HdrS":
                    raise ValueError(
                        "Vmlinuz file is not a Linux kernel bzImage."
                    )

                pref_address, init_size = struct.unpack(
                    "<QI", data[0x258:0x264]
                )
                self.logger.info(
                    "Vmlinuz pref_address is {:#x}, with init_size {:#x}."
                    .format(pref_address, init_size)
                )

                # Initramfs gets corrupted if it's too close to vmlinuz
                pad_to = align_up(data.size()) + small_pad

                # KASLR takes care to avoid overwriting initramfs in
                # self-decompression, but if that's disabled we need to
                # put the initramfs outside the decompression buffer.
                # But if the kernel and initramfs are small enough, they
                # might fit before pref_address, where we can skip that.
                low_usable = pref_address - align_up(data.size()) - small_pad
                if self.pad_vmlinuz and initramfs.stat().st_size > low_usable:
                    pad_to = align_up(addr_to_offs(pref_address + init_size))

                if pad_to > data.size():
                    self.logger.info(
                        "Padding vmlinuz to size {:#x}"
                        .format(pad_to)
                    )
                    data.resize(pad_to)

            # vbutil_kernel picks apart the vmlinuz in ways I don't
            # really want to reimplement right now, so just call it.
            self.logger.info("Packing files as temporary image.")
            temp_img = self._tempfile("temp.img")
            proc = vbutil_kernel(
                "--version", "1",
                "--arch", self.arch.vboot,
                "--vmlinuz", vmlinuz,
                "--config", cmdline_file,
                "--bootloader", initramfs,
                "--keyblock", self.keyblock,
                "--signprivate", self.signprivate,
                "--pack", temp_img,
            )
            self.logger.info(proc.stdout)

            # Do binary editing for now, until I get time to write
            # parsers for vboot_reference structs and kernel headers.
            with temp_img.open("r+b") as f, mmap(f.fileno(), 0) as data:
                if data[:8] != b"CHROMEOS":
                    raise RuntimeError(
                        "Unexpected output format from vbutil_kernel, "
                        "expected 'CHROMEOS' magic at start of file."
                    )

                # File starts with a keyblock and a kernel preamble
                # immediately afterwards, and padding up to 0x10000.
                keyblock_size = struct.unpack(
                    "<I", data[0x10:0x14]
                )[0]
                p = preamble_offset = keyblock_size

                # Preamble has the "memory address" of the "bootloader"
                # but it assumes the body is loaded at 0x100000.
                bootloader_addr = struct.unpack(
                    "<I", data[p+0x38:p+0x3c]
                )[0]
                bootloader_offset = addr_to_offs(bootloader_addr, 0x100000)

                # Assume vbutil_kernel correctly put it as "bootloader"
                initramfs_offset = bootloader_offset
                initramfs_addr = offs_to_addr(initramfs_offset)
                initramfs_size = initramfs.stat().st_size

                self.logger.info(
                    "Initramfs is at offset {:#x}, address {:#x}, size {:#x}."
                    .format(initramfs_offset, initramfs_addr, initramfs_size)
                )

                # Params is immediately before bootloader with size 0x1000
                p = params_offset = bootloader_offset - 0x1000
                if data[p+0x202:p+0x206] != b"HdrS":
                    raise RuntimeError(
                        "Unexpected output format from vbutil_kernel, "
                        "expected 'HdrS' magic in boot params."
                    )

                # These get passed to the kernel unmodified by depthcharge.
                # "initrdmem=addr,size" in the cmdline would work, but
                # this looks like how bootloaders are supposed to do it.
                data[p+0x218:p+0x21c] = struct.pack("<I", initramfs_addr)
                data[p+0x21c:p+0x220] = struct.pack("<I", initramfs_size)

                # Kernel self-decompression first copies vmlinuz to avoid
                # overwriting itself, ending at pref_address + init_size
                # or so. Increasing init_size in the boot params makes
                # it avoid overwriting our initramfs. Also needs some
                # small padding for unknown reasons.
                if self.set_init_size:
                    safe_copy_end = align_up(
                        initramfs_addr + initramfs_size
                        + small_pad
                        + vmlinuz.stat().st_size
                    )
                    if pref_address + init_size < safe_copy_end:
                        init_size = safe_copy_end - pref_address
                        self.logger.info(
                            "Setting init_size = {:#x}."
                            .format(init_size)
                        )
                        data[p+0x260:p+0x264] = struct.pack("<I", init_size)

            self.logger.info("Re-signing edited temporary image.")
            proc = vbutil_kernel(
                "--keyblock", self.keyblock,
                "--signprivate", self.signprivate,
                "--oldblob", temp_img,
                "--repack", self.output,
            )
            self.logger.info(proc.stdout)

        self.logger.info("Verifying built depthcharge image:")
        signpubkey_args = []
        if self.signpubkey is not None:
            signpubkey_args += ["--signpubkey", self.signpubkey]

        proc = vbutil_kernel(
            "--verify", self.output,
            *signpubkey_args,
        )
        self.logger.info(proc.stdout)

        return self.output


if __name__ == "__main__":
    mkdepthcharge.main()



#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools pathlib utilities
# Copyright (C) 2020-2022 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

import shutil
import subprocess

from pathlib import Path

from depthcharge_tools.utils.subprocess import (
    gzip,
    lz4,
    lzma,
    lzop,
    bzip2,
    xz,
    zstd,
)

def copy(src, dest):
    dest = shutil.copy2(src, dest)
    return Path(dest)


def decompress(src, dest=None, partial=False):
    if dest is not None:
        dest = Path(dest)

    for runner in (gzip, zstd, xz, lz4, lzma, bzip2, lzop):
        try:
            return runner.decompress(src, dest)

        except FileNotFoundError:
            if dest:
                dest.unlink()

        except subprocess.CalledProcessError as err:
            if dest is None and err.output and partial:
                return err.output

            elif dest and dest.stat().st_size > 0 and partial:
                return dest

            elif dest:
                dest.unlink()


def iterdir(path):
    try:
        if path.is_dir():
            return path.iterdir()
        else:
            return []
    except:
        return []


def read_lines(path):
    try:
        if path.is_file():
            return path.read_text().splitlines()
        else:
            return []
    except:
        return []



#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools platform utilities
# Copyright (C) 2020-2022 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

import collections
import glob
import platform
import re
import shlex

from pathlib import Path

from depthcharge_tools.utils.pathlib import (
    decompress,
)
from depthcharge_tools.utils.subprocess import (
    crossystem,
)


def dt_compatibles():
    dt_model = Path("/proc/device-tree/compatible")
    if dt_model.exists():
        return dt_model.read_text().strip("\x00").split("\x00")


def dt_model():
    dt_model = Path("/proc/device-tree/model")
    if dt_model.exists():
        return dt_model.read_text().strip("\x00")


def cros_hwid():
    hwid_file = Path("/proc/device-tree/firmware/chromeos/hardware-id")
    if hwid_file.exists():
        return hwid_file.read_text().strip("\x00")

    for hwid_file in Path("/sys/bus/platform/devices").glob("GGL0001:*/HWID"):
        if hwid_file.exists():
            return hwid_file.read_text().strip()

    # Try crossystem as a last resort
    try:
        return crossystem.hwid()
    except:
        pass


def cros_fwid():
    fwid_file = Path("/proc/device-tree/firmware/chromeos/firmware-version")
    if fwid_file.exists():
        return fwid_file.read_text().strip("\x00")

    for fwid_file in Path("/sys/bus/platform/devices").glob("GGL0001:*/FWID"):
        if fwid_file.exists():
            return fwid_file.read_text().strip()

    # Try crossystem as a last resort
    try:
        return crossystem.fwid()
    except:
        pass


def os_release(root=None):
    os_release = {}

    if root is None:
        root = "/"
    root = Path(root).resolve()

    os_release_f = root / "etc" / "os-release"
    if not os_release_f.exists():
        os_release_f = root / "usr" / "lib" / "os-release"

    if os_release_f.exists():
        for line in os_release_f.read_text().splitlines():
            lhs, _, rhs = line.partition("=")
            os_release[lhs] = rhs.strip('\'"')

    return os_release


def kernel_cmdline(root=None):
    cmdline = ""

    if root is None:
        root = "/"
    root = Path(root).resolve()

    cmdline_f = root / "etc" / "kernel" / "cmdline"
    if not cmdline_f.exists():
        cmdline_f = root / "usr" / "lib" / "kernel" / "cmdline"

    if cmdline_f.exists():
        cmdline = cmdline_f.read_text().rstrip("\n")

    return shlex.split(cmdline)


def proc_cmdline():
    cmdline = ""

    cmdline_f = Path("/proc/cmdline")
    if cmdline_f.exists():
        cmdline = cmdline_f.read_text().rstrip("\n")

    return shlex.split(cmdline)


def is_cros_boot():
    dt_cros_firmware = Path("/proc/device-tree/firmware/chromeos")
    if dt_cros_firmware.is_dir():
        return True

    # Chrome OS firmware injects this into the kernel cmdline.
    if "cros_secure" in proc_cmdline():
        return True

    return False


def is_cros_libreboot():
    fwid = cros_fwid()
    if fwid is None:
        return False

    return fwid.lower().startswith("libreboot")


def root_requires_initramfs(root):
    x = "[0-9a-fA-F]"
    uuid = "{x}{{8}}-{x}{{4}}-{x}{{4}}-{x}{{4}}-{x}{{12}}".format(x=x)
    ntsig = "{x}{{8}}-{x}{{2}}".format(x=x)

    # Depthcharge replaces %U with an uuid, so we can use that as well.
    uuid = "({}|%U)".format(uuid)

    # Tries to validate the root=* kernel cmdline parameter.
    # See init/do_mounts.c in Linux tree.
    for pat in (
        "[0-9a-fA-F]{4}",
        "/dev/nfs",
        "/dev/[0-9a-zA-Z]+",
        "/dev/[0-9a-zA-Z]+[0-9]+",
        "/dev/[0-9a-zA-Z]+p[0-9]+",
        "PARTUUID=({uuid}|{ntsig})".format(uuid=uuid, ntsig=ntsig),
        "PARTUUID=({uuid}|{ntsig})/PARTNROFF=[0-9]+".format(
            uuid=uuid, ntsig=ntsig,
        ),
        "[0-9]+:[0-9]+",
        "PARTLABEL=.+",
        "/dev/cifs",
    ):
        if re.fullmatch(pat, root):
            return False

    return True


def vboot_keys(*keydirs, system=True, root=None):
    if len(keydirs) == 0 or system:
        if root is None:
            root = "/"
        root = Path(root).resolve()

        keydirs = (
            *keydirs,
            root / "etc" / "depthcharge-tools",
            root / "usr" / "share" / "vboot" / "devkeys",
            root / "usr" / "local" / "share" / "vboot" / "devkeys",
        )

    for keydir in keydirs:
        keydir = Path(keydir)
        if not keydir.is_dir():
            continue

        keyblock = keydir / "kernel.keyblock"
        signprivate = keydir / "kernel_data_key.vbprivk"
        signpubkey = keydir / "kernel_subkey.vbpubk"

        if not keyblock.exists():
            keyblock = None
        if not signprivate.exists():
            signprivate = None
        if not signpubkey.exists():
            signpubkey = None

        if keyblock or signprivate or signpubkey:
            return keydir, keyblock, signprivate, signpubkey

    return None, None, None, None


def cpu_microcode(boot=None):
    microcode = []

    for f in (
        *boot.glob("amd-ucode.img"),
        *boot.glob("amd-uc.img"),
    ):
        if f.is_file():
            microcode.append(f)
            break

    for f in (
        *boot.glob("intel-ucode.img"),
        *boot.glob("intel-uc.img"),
    ):
        if f.is_file():
            microcode.append(f)
            break

    if not microcode:
        for f in (
            *boot.glob("early_ucode.cpio"),
            *boot.glob("microcode.cpio"),
        ):
            if f.is_file():
                microcode.append(f)
                break

    return microcode


def installed_kernels(root=None, boot=None):
    kernels = {}
    initrds = {}
    fdtdirs = {}

    if root is None:
        root = "/"
    root = Path(root).resolve()

    if boot is None:
        boot = root / "boot"
    boot = Path(boot).resolve()

    for f in (
        *root.glob("lib/modules/*/vmlinuz"),
        *root.glob("lib/modules/*/vmlinux"),
        *root.glob("lib/modules/*/Image"),
        *root.glob("lib/modules/*/zImage"),
        *root.glob("lib/modules/*/bzImage"),
        *root.glob("usr/lib/modules/*/vmlinuz"),
        *root.glob("usr/lib/modules/*/vmlinux"),
        *root.glob("usr/lib/modules/*/Image"),
        *root.glob("usr/lib/modules/*/zImage"),
        *root.glob("usr/lib/modules/*/bzImage"),
    ):
        if not f.is_file():
            continue
        release = f.parent.name
        kernels[release] = f.resolve()

    for f in (
        *boot.glob("vmlinuz-*"),
        *boot.glob("vmlinux-*"),
    ):
        if not f.is_file():
            continue
        _, _, release = f.name.partition("-")
        kernels[release] = f.resolve()

    for f in (
        *boot.glob("vmlinuz"),
        *boot.glob("vmlinux"),
        *root.glob("vmlinuz"),
        *root.glob("vmlinux"),
        *boot.glob("Image"),
        *boot.glob("zImage"),
        *boot.glob("bzImage"),
    ):
        if not f.is_file():
            continue
        kernels[None] = f.resolve()
        break

    for f in (
        *root.glob("lib/modules/*/initrd"),
        *root.glob("lib/modules/*/initramfs"),
        *root.glob("lib/modules/*/initrd.img"),
        *root.glob("lib/modules/*/initramfs.img"),
        *root.glob("usr/lib/modules/*/initrd"),
        *root.glob("usr/lib/modules/*/initramfs"),
        *root.glob("usr/lib/modules/*/initrd.img"),
        *root.glob("usr/lib/modules/*/initramfs.img"),
    ):
        if not f.is_file():
            continue
        release = f.parent.name
        initrds[release] = f.resolve()

    for f in (
        *boot.glob("initrd-*.img"),
        *boot.glob("initramfs-*.img"),
    ):
        if not f.is_file():
            continue
        _, _, release = f.name.partition("-")
        release = release[:-4]
        initrds[release] = f.resolve()

    for f in (
        *boot.glob("initrd-*"),
        *boot.glob("initrd.img-*"),
        *boot.glob("initramfs-*"),
        *boot.glob("initramfs.img-*"),
    ):
        if not f.is_file():
            continue
        _, _, release = f.name.partition("-")
        initrds[release] = f.resolve()

    for f in (
        *boot.glob("initrd.img"),
        *boot.glob("initrd"),
        *boot.glob("initramfs-linux.img"),
        *boot.glob("initramfs-vanilla"),
        *boot.glob("initramfs"),
        *root.glob("initrd.img"),
        *root.glob("initrd"),
        *root.glob("initramfs"),
    ):
        if not f.is_file():
            continue
        initrds[None] = f.resolve()
        break

    for d in (
        *root.glob("usr/lib/linux-image-*"),
    ):
        if not d.is_dir():
            continue
        _, _, release = d.name.partition("linux-image-")
        fdtdirs[release] = d.resolve()

    for d in (
        *root.glob("lib/modules/*/dtb"),
        *root.glob("lib/modules/*/dtbs"),
        *root.glob("usr/lib/modules/*/dtb"),
        *root.glob("usr/lib/modules/*/dtbs"),
    ):
        if not d.is_dir():
            continue
        release = d.parent.name
        fdtdirs[release] = d.resolve()

    for d in (
        *boot.glob("dtb-*"),
        *boot.glob("dtbs-*"),
    ):
        if not d.is_dir():
            continue
        _, _, release = d.name.partition("-")
        fdtdirs[release] = d.resolve()

    for d in (
        *boot.glob("dtb/*"),
        *boot.glob("dtbs/*"),
    ):
        if not d.is_dir():
            continue
        if d.name in kernels:
            fdtdirs[d.name] = d.resolve()

    for d in (
        *boot.glob("dtbs"),
        *boot.glob("dtb"),
        *root.glob("usr/share/dtbs"),
        *root.glob("usr/share/dtb"),
    ):
        if not d.is_dir():
            continue
        # Duplicate dtb files means that the directory is split by
        # kernel release and we can't use it for a single release.
        dtbs = d.glob("**/*.dtb")
        counts = collections.Counter(dtb.name for dtb in dtbs)
        if all(c <= 1 for c in counts.values()):
            fdtdirs[None] = d.resolve()
            break

    if None in kernels:
        kernel, release = kernels[None], None
        for r, k in kernels.items():
            if k == kernel and r is not None:
                release = r
                break

        if release is not None:
            del kernels[None]
            if None in initrds:
                initrds.setdefault(release, initrds[None])
                del initrds[None]
            if None in fdtdirs:
                fdtdirs.setdefault(release, fdtdirs[None])
                del fdtdirs[None]

    return [
        KernelEntry(
            release,
            kernel=kernels[release],
            initrd=initrds.get(release, None),
            fdtdir=fdtdirs.get(release, None),
            os_name=os_release(root=root).get("NAME", None),
        ) for release in kernels.keys()
    ]


class KernelEntry:
    def __init__(self, release, kernel, initrd=None, fdtdir=None, os_name=None):
        self.release = release
        self.kernel = kernel
        self.initrd = initrd
        self.fdtdir = fdtdir
        self.os_name = os_name

    @property
    def description(self):
        if self.os_name is None:
            return "Linux {}".format(self.release)
        else:
            return "{}, with Linux {}".format(self.os_name, self.release)

    @property
    def arch(self):
        kernel = Path(self.kernel)

        decomp = decompress(kernel)
        if decomp:
            head = decomp[:4096]
        else:
            with kernel.open("rb") as f:
                head = f.read(4096)

        if head[0x202:0x206] == b"HdrS":
            return Architecture("x86")
        elif head[0x38:0x3c] == b"ARM\x64":
            return Architecture("arm64")
        elif head[0x34:0x38] == b"\x45\x45\x45\x45":
            return Architecture("arm")

    def _comparable_parts(self):
        pattern = "([^a-zA-Z0-9]?)([a-zA-Z]*)([0-9]*)"

        if self.release is None:
            return ()

        parts = []
        for sep, text, num in re.findall(pattern, self.release):
            # x.y.z > x.y-* == x.y* > x.y~*
            sep = {
                "~": -1,
                ".": 1,
            }.get(sep, 0)

            # x.y-* == x.y* > x.y > x.y-rc* == x.y-trunk*
            text = ({
                "rc": -1,
                "trunk": -1,
            }.get(text, 0), text)

            # Compare numbers as numbers
            num = int(num) if num else 0

            parts.append((sep, text, num))

        return tuple(parts)

    def __lt__(self, other):
        if not isinstance(other, KernelEntry):
            return NotImplemented

        return self._comparable_parts() < other._comparable_parts()

    def __gt__(self, other):
        if not isinstance(other, KernelEntry):
            return NotImplemented

        return self._comparable_parts() > other._comparable_parts()

    def __str__(self):
        return self.description

    def __repr__(self):
        return (
            "KernelEntry(release={!r}, kernel={!r}, initrd={!r}, fdtdir={!r}, os_name={!r})"
            .format(self.release, self.kernel, self.initrd, self.fdtdir, self.os_name)
        )


class Architecture(str):
    arm_32 = ["arm", "ARM", "armv7", "ARMv7", ]
    arm_64 = ["arm64", "ARM64", "aarch64", "AArch64"]
    arm = arm_32 + arm_64
    x86_32 = ["i386", "x86"]
    x86_64 = ["x86_64", "amd64", "AMD64"]
    x86 = x86_32 + x86_64
    all = arm + x86
    groups = (arm_32, arm_64, x86_32, x86_64)

    def __eq__(self, other):
        if isinstance(other, Architecture):
            for group in self.groups:
                if self in group and other in group:
                    return True
        return str(self) == str(other)

    def __ne__(self, other):
        if isinstance(other, Architecture):
            for group in self.groups:
                if self in group and other not in group:
                    return True
        return str(self) != str(other)

    @property
    def mkimage(self):
        if self in self.arm_32:
            return "arm"
        if self in self.arm_64:
            return "arm64"
        if self in self.x86_32:
            return "x86"
        if self in self.x86_64:
            return "x86_64"

    @property
    def vboot(self):
        if self in self.arm_32:
            return "arm"
        if self in self.arm_64:
            return "aarch64"
        if self in self.x86_32:
            return "x86"
        if self in self.x86_64:
            return "amd64"

    @property
    def kernel_arches(self):
        if self in self.arm_32:
            return self.arm_32
        if self in self.arm_64:
            return self.arm
        if self in self.x86_32:
            return self.x86_32
        if self in self.x86_64:
            return self.x86



#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools depthchargectl command
# Copyright (C) 2020-2023 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

import argparse
import collections
import configparser
import copy
import glob
import logging
import os
import platform
import re
import shlex
import tempfile

from pathlib import Path

from depthcharge_tools import (
    __version__,
    config_ini,
    boards_ini,
    config_files,
)
from depthcharge_tools.utils.argparse import (
    Command,
    Argument,
    Group,
    Subparsers,
)
from depthcharge_tools.utils.collections import (
    ConfigDict,
)
from depthcharge_tools.utils.os import (
    Disks,
)
from depthcharge_tools.utils.platform import (
    Architecture,
    vboot_keys,
    cros_hwid,
    dt_compatibles,
    is_cros_boot,
    is_cros_libreboot,
    kernel_cmdline,
    proc_cmdline,
)
from depthcharge_tools.utils.string import (
    parse_bytesize,
)


class Board:
    def __init__(self, config):
        self._config = config

    @property
    def name(self):
        name = self._config.get("name")
        if name is None:
            name = "Unnamed {} board".format(self.codename or 'unknown')
        return name

    @property
    def codename(self):
        return self._config.get("codename")

    @property
    def arch(self):
        return Architecture(self._config.get("arch"))

    @property
    def dt_compatible(self):
        pattern = self._config.get("dt-compatible")

        # Try to detect non-regex values and extend them to match any
        # rev/sku, but if a rev/sku is given match only the given one.
        if pattern and re.fullmatch("[\w,-]+", pattern):
            prefix, rev, sku = re.fullmatch(
                "(.*?)(-rev\d+)?(-sku\d+)?",
                pattern,
            ).groups()

            pattern = "{}{}{}".format(
                prefix,
                rev or "(-rev\d+)?",
                sku or "(-sku\d+)?",
            )

        if pattern:
            return re.compile(pattern)

    @property
    def hwid_match(self):
        pattern = self._config.get("hwid-match")
        if pattern in (None, "None", "none"):
            return None

        if pattern:
            return re.compile(pattern)

    @property
    def boots_lz4_kernel(self):
        return self._config.getboolean("boots-lz4-kernel", False)

    @property
    def boots_lzma_kernel(self):
        return self._config.getboolean("boots-lzma-kernel", False)

    @property
    def loads_zimage_ramdisk(self):
        return self._config.getboolean("loads-zimage-ramdisk", False)

    @property
    def loads_fit_ramdisk(self):
        return self._config.getboolean("loads-fit-ramdisk", False)

    @property
    def loads_dtb_off_by_one(self):
        return self._config.getboolean("loads-dtb-off-by-one", False)

    @property
    def fit_ramdisk_load_address(self):
        addr = self._config.get("fit-ramdisk-load-address", None)
        return parse_bytesize(addr)

    @property
    def image_start_address(self):
        addr = self._config.get("image-start-address", None)
        return parse_bytesize(addr)

    @property
    def image_max_size(self):
        max_size = self._config.get("image-max-size")
        if max_size in (None, "None", "none"):
            return float("inf")

        return parse_bytesize(max_size)

    @property
    def image_format(self):
        return self._config.get("image-format")


class depthchargectl(
    Command,
    prog="depthchargectl",
    usage="%(prog)s [options] command ...",
    add_help=False,
):
    """Manage Chrome OS kernel partitions."""

    _logger = logging.getLogger(__name__)
    config_section = "depthchargectl"

    @property
    def logger(self):
        # Set verbosity before logging messages
        self.verbosity
        return self._logger

    @Group
    def global_options(self):
        """Global options"""

    @global_options.add
    @Argument("-h", "--help", action="help")
    def print_help(self):
        """Show this help message."""
        # type(self).parser.print_help()

    @global_options.add
    @Argument(
        "-V", "--version",
        action="version",
        version="depthcharge-tools %(prog)s {}".format(__version__),
    )
    def version(self):
        """Print program version."""
        return type(self).version.version % {"prog": type(self).prog}

    @global_options.add
    @Argument("-v", "--verbose", count=True)
    def verbosity(self, verbosity=0):
        """Print more detailed output."""
        level = logging.WARNING - int(verbosity) * 10
        self._logger.setLevel(level)
        return verbosity

    @global_options.add
    @Argument("--tmpdir", nargs=1)
    def tmpdir(self, dir_=None):
        """Directory to keep temporary files."""
        if dir_ is None:
            dir_ = tempfile.TemporaryDirectory(
                prefix="depthchargectl-",
            )
            dir_ = self.exitstack.enter_context(dir_)

        dir_ = Path(dir_)
        os.makedirs(dir_, exist_ok=True)

        self.logger.debug("Working in temp dir '{}'.".format(dir_))

        return dir_

    @global_options.add
    @Argument("--root", nargs=1)
    def root(self, root=None):
        """Root device or mountpoint of the system to work on."""
        if root is None:
            cmdline = self.kernel_cmdline

            for c in cmdline:
                lhs, _, rhs = c.partition("=")
                if lhs.lower() == "root":
                    root = rhs

            if root:
                self.logger.info(
                    "Using root '{}' set in cmdline."
                    .format(root)
                )
                return str(root)

            self.logger.info(
                "Defaulting to current system root '/'."
            )
            return Path("/").resolve()

        if os.path.ismount(Path(root).resolve()):
            self.logger.info(
                "Using root argument '{}' as the system to work on."
                .format(root)
            )
            return Path(root).resolve()

        if root in (None, "", "none", "None"):
            self.logger.info(
                "Using no root argument for the kernel cmdline."
                .format(root)
            )
        else:
            self.logger.info(
                "Using root argument '{}' as a device description."
                .format(root)
            )

        return str(root)

    @global_options.add
    @Argument("--root-mountpoint", nargs=1, metavar="DIR")
    def root_mountpoint(self, mnt=None):
        """Root mountpoint of the system to work on."""
        if mnt:
            mnt = Path(mnt).resolve()
            self.logger.info(
                "Using root mountpoint '{}' from given argument."
                .format(mnt)
            )
            return mnt

        if self.root in ("", "None", "none", None):
            return Path("/").resolve()

        if isinstance(self.root, Path):
            return self.root

        disk = self.diskinfo.evaluate(self.root)
        mountpoints = sorted(
            self.diskinfo.mountpoints(disk),
            key=lambda p: len(p.parents),
        )

        if len(mountpoints) > 1:
            mnt = mountpoints[0]
            self.logger.warning(
                "Choosing '{}' from multiple root mountpoints: {}."
                .format(mnt, ", ".join(str(m) for m in mountpoints))
            )
            return mnt

        elif mountpoints:
            mnt = mountpoints[0]
            if mnt != Path("/").resolve():
                self.logger.info(
                    "Using root mountpoint '{}'."
                    .format(mnt)
                )
            return mnt

        self.logger.warning(
            "Couldn't find root mountpoint, falling back to '/'."
        )

        return Path("/").resolve()

    @global_options.add
    @Argument("--boot-mountpoint", nargs=1, metavar="DIR")
    def boot_mountpoint(self, boot=None):
        """Boot mountpoint of the system to work on."""
        if boot:
            boot = Path(boot).resolve()
            self.logger.info(
                "Using boot mountpoint '{}' from given argument."
                .format(boot)
            )
            return boot

        boot_str = self.diskinfo.by_mountpoint("/boot", fstab_only=True)
        device = self.diskinfo.evaluate(boot_str)
        mountpoints = sorted(
            self.diskinfo.mountpoints(device),
            key=lambda p: len(p.parents),
        )

        if device and not mountpoints:
            self.logger.warning(
                "Boot partition '{}' for specified root is not mounted."
                .format(device)
            )

        if len(mountpoints) > 1:
            self.logger.warning(
                "Choosing '{}' from multiple /boot mountpoints: {}."
                .format(mountpoints[0], ", ".join(str(m) for m in mountpoints))
            )

        if mountpoints:
            return mountpoints[0]

        root = self.root_mountpoint
        boot = (root / "boot").resolve()
        self.logger.info(
            "Couldn't find /boot in fstab, falling back to '{}'."
            .format(boot)
        )

        if root != Path("/").resolve() and not boot.is_dir():
            self.logger.warning(
                "Boot mountpoint '{}' does not exist for custom root."
                .format(boot)
            )
            self.logger.warning(
                "Not falling back to the running system for boot mountpoint."
            )

        return boot

    @Argument(dest=argparse.SUPPRESS, help=argparse.SUPPRESS)
    def diskinfo(self):
        # Break cyclic dependencies here
        yield Disks()

        root = self.root_mountpoint

        return Disks(
            fstab=(root / "etc" / "fstab"),
            crypttab=(root / "etc" / "crypttab"),
        )

    @Group
    def config_options(self):
        """Configuration options"""

        # Update the values in the configparser object so that the
        # config subcommand can query e.g. the autodetected board.
        self.config.update({
            'board': self.board.codename if self.board else "none",
            'images-dir': str(self.images_dir),
            'vboot-keyblock': str(self.vboot_keyblock),
            'vboot-public-key': str(self.vboot_public_key),
            'vboot-private-key': str(self.vboot_private_key),
            'kernel-cmdline': " ".join(self.kernel_cmdline),
            'ignore-initramfs': str(self.ignore_initramfs),
            'zimage-initramfs-hack': str(self.zimage_initramfs_hack),
        })

        if self.board is not None:
            def pattern(regex):
                try:
                    return str(regex.pattern)
                except:
                    return "None"

            self.config.update({
                "arch": str(self.board.arch),
                "codename": str(self.board.codename),
                "boots-lz4-kernel": str(self.board.boots_lz4_kernel),
                "boots-lzma-kernel": str(self.board.boots_lzma_kernel),
                "dt-compatible": pattern(self.board.dt_compatible),
                "fit-ramdisk-load-address": str(self.board.fit_ramdisk_load_address),
                "hwid-match": pattern(self.board.hwid_match),
                "image-format": str(self.board.image_format),
                "image-max-size": str(self.board.image_max_size),
                "image-start-address": str(self.board.image_start_address),
                "loads-dtb-off-by-one": str(self.board.loads_dtb_off_by_one),
                "loads-fit-ramdisk": str(self.board.loads_fit_ramdisk),
                "loads-zimage-ramdisk": str(self.board.loads_zimage_ramdisk),
                "name": str(self.board.name),
            })

    @config_options.add
    @Argument("--config", nargs=1)
    def config(self, file_=None):
        """Additional configuration file to read"""
        if isinstance(file_, configparser.SectionProxy):
            parser = file_.parser

        elif isinstance(file_, configparser.ConfigParser):
            parser = file_
            file_ = None

        else:
            parser = configparser.ConfigParser(
                default_section="depthcharge-tools",
                dict_type=ConfigDict,
            )

            parser.read_string(config_ini, source="config.ini")
            parser.read_string(boards_ini, source="boards.ini")

            try:
                for p in parser.read(config_files):
                    self.logger.debug("Read config file '{}'.".format(p))

            except configparser.ParsingError as err:
                self.logger.warning(
                    "Config file '{}' could not be parsed."
                    .format(err.filename)
                )

        root = self.root_mountpoint
        boot = self.boot_mountpoint

        def fixup_path(f):
            p = Path(f).resolve()
            if f.startswith("/boot"):
                return str(boot / p.relative_to("/boot"))
            elif f.startswith("/"):
                return str(root / p.relative_to("/"))

        if root != Path("/").resolve():
            extra_parser = configparser.ConfigParser()
            extra_files = [
                *root.glob("etc/depthcharge-tools/config"),
                *root.glob("etc/depthcharge-tools/config.d/*"),
            ]

            try:
                for p in extra_parser.read(extra_files):
                    self.logger.debug("Read config file '{}'.".format(p))

            except configparser.ParsingError as err:
                self.logger.warning(
                    "Config file '{}' could not be parsed."
                    .format(err.filename)
                )

            file_configs = [
                "images-dir",
                "vboot-keyblock",
                "vboot-public-key",
                "vboot-private-key",
            ]

            for sect, conf in extra_parser.items():
                for key, value in conf.items():
                    if key in file_configs:
                        conf[key] = fixup_path(value)

            parser.read_dict(extra_parser)

        if self.config_section not in parser.sections():
            if self.config_section != parser.default_section:
                parser.add_section(self.config_section)

        if isinstance(file_, collections.abc.Mapping):
            parser[self.config_section].update(file_)

        elif file_ is not None:
            try:
                read = parser.read([file_])

            except configparser.ParsingError as err:
                raise ValueError(
                    "Config file '{}' could not be parsed."
                    .format(err.filename)
                )

            if file_ not in read:
                raise ValueError(
                    "Config file '{}' could not be read."
                    .format(file_)
                )

        return parser[self.config_section]

    @config_options.add
    @Argument("--board", nargs=1)
    def board(self, codename=""):
        """Assume we're running on the specified board"""
        if isinstance(codename, Board):
            return codename

        elif isinstance(codename, configparser.SectionProxy):
            return Board(codename)

        elif codename is None:
            return None

        boards = {
            sectname: Board(section)
            for sectname, section in self.config.parser.items()
            if sectname.startswith("boards/")
        }

        if not codename:
            codename = self.config.get("board", "")

        if codename in ("None", "none"):
            return None

        elif codename:
            parts = str(codename).lower().replace('-', '_').split('_')

            def codename_match(item):
                if item is None:
                    return (len(parts) - 1, 0)

                sectname, board = item
                matchparts = sectname.split("/") + (
                    [] if board.codename is None else
                    board.codename.lower().replace('-', '_').split('_')
                )

                # Don't match sections without explicit codenames
                parent, _, _ = sectname.rpartition('/')
                if parent in boards and boards[parent].codename == board.codename:
                    return (len(parts) - 1, float("inf"))

                # Some kind of a fuzzy match, how many parts of the
                # given codename exist in the parts of this config
                idx = len(parts) - 1
                while parts and matchparts and idx >= 0:
                    if parts[idx] == matchparts[-1]:
                        idx -= 1
                    # Oldest boards have x86-alex_he etc.
                    elif (parts[idx], matchparts[-1]) == ("x86", "amd64"):
                        idx -= 1
                    matchparts.pop()

                # Avoid matching only on "libreboot" without actual board
                if parts[-1] == "libreboot" and idx == len(parts) - 2:
                    return (len(parts) - 1, float("inf"))

                return (idx, len(sectname.split("/")))

            match_groups = collections.defaultdict(list)
            for item in (None, *boards.items()):
                match_groups[codename_match(item)].append(item)

            score, matches = min(match_groups.items())
            if not matches or None in matches:
                raise ValueError(
                    "Unknown board codename '{}'."
                    .format(codename)
                )

            elif len(matches) > 1:
                raise ValueError(
                    "Ambiguous board codename '{}' matches {}."
                    .format(codename, [b.codename for s, b in matches])
                )

            sectname, board = matches[0]
            self.logger.info(
                "Assuming board '{}' ('{}') by codename argument or config."
                .format(board.name, board.codename)
            )

            return board

        hwid = cros_hwid()
        def hwid_match(item):
            sectname, board = item
            try:
                return bool(re.match(board.hwid_match, hwid))
            except:
                return False

        if hwid is not None:
            matches = tuple(filter(hwid_match, boards.items()))
        else:
            matches = ()

        if matches:
            sectname, board = matches[0]

            if is_cros_libreboot():
                libreboot_name = "{}/libreboot".format(sectname)
                libreboot_board = boards.get(libreboot_name, None)
                if libreboot_board:
                    sectname = libreboot_name
                    board = libreboot_board

            self.logger.info(
                "Detected board '{}' ('{}') by HWID."
                .format(board.name, board.codename)
            )
            return board

        compatibles = dt_compatibles()
        def compat_preference(item):
            if item is None:
                return (len(compatibles), 0)

            sectname, board = item
            if board.dt_compatible is None:
                return (float("inf"), 0)

            for i, c in enumerate(compatibles):
                if board.dt_compatible.fullmatch(c):
                    return (i, -len(sectname.split("/")))
            else:
                return (float("inf"), -1)

        if compatibles is not None:
            match = min((None, *boards.items()), key=compat_preference)
        else:
            match = None

        if match is not None:
            sectname, board = match

            if is_cros_libreboot():
                libreboot_name = "{}/libreboot".format(sectname)
                libreboot_board = boards.get(libreboot_name, None)
                if libreboot_board:
                    sectname = libreboot_name
                    board = libreboot_board

            self.logger.info(
                "Detected board '{}' ('{}') by device-tree compatibles."
                .format(board.name, board.codename)
            )
            return board

        # This might actually be running on non-ChromeOS hardware.
        # Check this after the board detection code, because we might
        # also be running on e.g. RW_LEGACY but still with depthcharge.
        if not is_cros_boot():
            return None

        # Use generic boards per cpu architecture, since we couldn't
        # detect this system as a proper board
        arch = platform.machine()
        if arch in Architecture.arm_32:
            sectname = "boards/arm"
        elif arch in Architecture.arm_64:
            sectname = "boards/arm64"
        elif arch in Architecture.x86:
            sectname = "boards/amd64"
        board = boards.get(sectname, None)
        if board is not None:
            self.logger.warning(
                "Assuming a generic board of architecture '{}'."
                .format(board.arch)
            )
            return board

        raise ValueError(
            "Could not detect which board this is running on."
        )

    @config_options.add
    @Argument("--images-dir", nargs=1)
    def images_dir(self, dir_=None):
        """Directory to store built images"""
        if dir_ is not None:
            return Path(dir_).resolve()

        dir_ = self.config.get("images-dir")
        if dir_ is None:
            raise ValueError(
                "Images directory is not specified"
            )

        dir_ = Path(dir_).resolve()

        boot = self.boot_mountpoint
        if boot != Path("/boot").resolve() and dir_.is_relative_to("/boot"):
            return boot / dir_.relative_to("/boot")

        root = self.root_mountpoint
        if root != Path("/").resolve() and not dir_.is_relative_to(root):
            return root / dir_.relative_to("/")

        return dir_

    @config_options.add
    @Argument("--vboot-keydir", nargs=1, help=argparse.SUPPRESS)
    def vboot_keydir(self, dir_=None):
        """Directory containing default vboot keys to use"""
        if dir_:
            keydir = vboot_keys(dir_, system=None)[0]
            if keydir:
                return Path(keydir).resolve()

        root = self.root_mountpoint
        if root != Path("/").resolve():
            keydir = vboot_keys(root=root)[0]
            if keydir:
                return Path(keydir).resolve()

        keydir = vboot_keys()[0]
        if keydir:
            return Path(keydir).resolve()

        return None

    @config_options.add
    @Argument("--vboot-keyblock", nargs=1)
    def vboot_keyblock(self, keyblock=None):
        """Keyblock file to include in images"""
        if keyblock:
            return Path(keyblock).resolve()

        keyblock = self.config.get("vboot-keyblock")
        if keyblock:
            return Path(keyblock).resolve()

        if self.vboot_keydir:
            keyblock = self.vboot_keydir / "kernel.keyblock"
            if keyblock.exists():
                return Path(keyblock).resolve()

    @config_options.add
    @Argument("--vboot-public-key", nargs=1)
    def vboot_public_key(self, signpubkey=None):
        """Public key file to verify images with"""
        if signpubkey:
            return Path(signpubkey).resolve()

        signpubkey = self.config.get("vboot-public-key")
        if signpubkey:
            return Path(signpubkey).resolve()

        if self.vboot_keydir:
            signpubkey = self.vboot_keydir / "kernel_subkey.vbpubk"
            if signpubkey.exists():
                return Path(signpubkey).resolve()

        return signpubkey

    @config_options.add
    @Argument("--vboot-private-key", nargs=1)
    def vboot_private_key(self, signprivate=None):
        """Private key file to sign images with"""
        if signprivate:
            return Path(signprivate).resolve()

        signprivate = self.config.get("vboot-private-key")
        if signprivate:
            return Path(signprivate).resolve()

        if self.vboot_keydir:
            signprivate = self.vboot_keydir / "kernel_data_key.vbprivk"
            if signprivate.exists():
                return Path(signprivate).resolve()

        return signprivate

    @config_options.add
    @Argument("--kernel-cmdline", nargs="+", metavar="CMD")
    def kernel_cmdline(self, *cmds):
        """Command line options for the kernel"""
        # Break cyclic dependencies here
        yield []
        cmdline_src = "given options"

        if len(cmds) == 0:
            cmdline = self.config.get("kernel-cmdline")
            if cmdline is not None:
                cmds = shlex.split(cmdline)
                cmdline_src = "configuration"

        if len(cmds) == 0:
            cmds = kernel_cmdline(self.root_mountpoint)
            cmdline_src = "/etc/kernel/cmdline"

        if len(cmds) == 0:
            if self.root_mountpoint == Path("/").resolve():
                cmds = [
                    cmd for cmd in proc_cmdline()
                    if cmd.split("=", 1)[0] not in (
                        "root", "initrd", "kern_guid", "BOOT_IMAGE",
                        "cros_secure", "cros_legacy", "cros_efi",
                    )
                ]
                cmdline_src = "/proc/cmdline"

        flat_cmds = []
        for cmd in cmds:
            flat_cmds.extend(shlex.split(cmd))

        if flat_cmds:
            self.logger.info(
                "Using kernel cmdline from {}: {}"
                .format(cmdline_src, " ".join(flat_cmds))
            )

        return flat_cmds

    @config_options.add
    @Argument("--ignore-initramfs", ignore=True)
    def ignore_initramfs(self, ignore=None):
        """Do not include initramfs in images"""
        if ignore is None:
            ignore = self.config.getboolean("ignore-initramfs", False)

        return ignore

    @config_options.add
    @Argument("--zimage-initramfs-hack", nargs=1, help=argparse.SUPPRESS)
    def zimage_initramfs_hack(self, hack=None):
        """Which initramfs support hack should be used for zimage format"""
        if hack is None:
            hack = self.config.get("zimage-initramfs-hack", "init-size")

        if hack in ("None", "none"):
            hack = None

        return hack

    @Subparsers()
    def command(self, cmd):
        """Supported subcommands"""

    def __call__(self):
        if hasattr(type(self), "list"):
            self.logger.info("No subcommand given, defaulting to list")
            return type(self).list()
        else:
            raise ValueError("No subcommand given")


import depthcharge_tools.depthchargectl._bless
import depthcharge_tools.depthchargectl._build
import depthcharge_tools.depthchargectl._check
import depthcharge_tools.depthchargectl._config
import depthcharge_tools.depthchargectl._list
import depthcharge_tools.depthchargectl._remove
import depthcharge_tools.depthchargectl._target
import depthcharge_tools.depthchargectl._write



#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools argparse utilities
# Copyright (C) 2020-2022 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

import argparse
import copy
import contextlib
import functools
import inspect
import logging
import sys


def filter_action_kwargs(kwargs, action="store"):
    """
    Filter out the kwargs which argparse actions don't recognize.

    ArgumentParser.add_argument() raises an error on unknown kwargs,
    filter them out. Also unset any values that are None.
    """

    action = kwargs.get("action", action)
    allowed = {
        "action", "dest", "nargs", "const", "default", "type",
        "choices", "required", "help", "metavar",
    }

    if action == "store":
        pass

    elif action == "store_const":
        allowed -= {"nargs", "type", "choices"}

    elif action == "store_true":
        allowed -= {"nargs", "const", "type", "choices", "metavar"}

    elif action == "store_false":
        allowed -= {"nargs", "const", "type", "choices", "metavar"}

    elif action == "append":
        pass

    elif action == "append_const":
        allowed -= {"nargs", "type", "choices"}

    elif action == "count":
        allowed -= {"nargs", "const", "type", "choices", "metavar"}

    elif action == "help":
        allowed = {"action", "dest", "default", "help"}

    elif action == "version":
        allowed = {"action", "version", "dest", "default", "help"}

    elif action is FunctionBindAction:
        allowed |= {"func", "append", "count", "args", "kwargs"}

    else:
        allowed = kwargs.keys()

    action_kwargs = {}
    other_kwargs = {}

    for key, value in kwargs.items():
        if key in allowed:
            action_kwargs[key] = value
        else:
            other_kwargs[key] = value

    return action_kwargs, other_kwargs


class FunctionBindAction(argparse.Action):
    def __init__(
        self,
        option_strings,
        dest,
        func,
        append=False,
        count=False,
        **kwargs,
    ):
        self.signature = inspect.signature(func)
        self.f_args = kwargs.pop("args", ())
        self.f_kwargs = kwargs.pop("kwargs", {})
        self.append = append
        self.count = count

        if append and kwargs.get("nargs", "*") == 0:
            raise ValueError(
                "'{}' action '{}' with append=True must be able to "
                "consume command-line arguments (nargs must not be 0)"
                .format(type(self).__name__, dest)
            )

        if count and kwargs.get("nargs", 0) != 0:
            raise ValueError(
                "'{}' action '{}' with count=True can't consume any "
                "command-line arguments (nargs must be 0)"
                .format(type(self).__name__, dest)
            )

        if count and append:
            raise ValueError(
                "'{}' action '{}' arguments append=True and count=True "
                "are incompatible."
                .format(type(self).__name__, dest)
            )

        if (count or append) and (self.f_args or self.f_kwargs):
            raise NotImplementedError(
                "'{}' action '{}' with append=True or count=True "
                "does not support prebinding arguments yet."
            )

        super_kwargs, _ = filter_action_kwargs(kwargs)
        super().__init__(option_strings, dest, **super_kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        if self.dest in (None, argparse.SUPPRESS):
            return

        current = getattr(namespace, self.dest, None)

        if self.nargs in ("?", None):
            values = [values]

        if self.append:
            args = current.args if current else ()
            bound = self.signature.bind_partial(*args, *values)

        elif self.count:
            n = int(current.args[0]) if current else 0
            bound = self.signature.bind(n + 1)

        else:
            bound = self.signature.bind(
                *self.f_args,
                *values,
                **self.f_kwargs,
            )

        setattr(namespace, self.dest, bound)


class _MethodDecorator:
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.__name__ = None
        self.__self__ = None
        self.__func__ = None

        if args and callable(args[0]):
            self.wrap(args[0])
            args = args[1:]

        self._args = args
        self._kwargs = kwargs

    def __get__(self, instance, owner):
        if self.__self__ is not None:
            return self

        if instance is None:
            bound = copy.copy(self)
            bound.__self__ = owner
            return bound

        if self.__name__ not in instance.__dict__:
            if self.__func__ is not None:
                return self.__func__.__get__(instance, owner)
            return None

        return instance.__dict__[self.__name__]

    def __set__(self, instance, value):
        instance.__dict__[self.__name__] = value

    def __set_name__(self, owner, name):
        self.__name__ = name

    @property
    def __call__(self):
        if self.__self__ is None:
            return self.wrap

        if self.__func__ is None:
            return None

        call = self.__func__.__get__(self, type(self))
        self.__signature__ = inspect.signature(call)

        return call

    def wrap(self, func):
        if not callable(func):
            raise TypeError("Can't wrap non-callable objects")

        self.__func__ = func
        functools.update_wrapper(self, func)
        self.__name__ = func.__name__

        return self

    def copy(self, *args, **kwargs):
        # Non-call decorator form
        if len(args) == 1 and callable(args[0]) and not kwargs:
            func, *args = args
        else:
            func = None

        args = (*self._args, *args)
        kwargs = {**self._kwargs, **kwargs}
        obj = type(self)(*args, **kwargs)
        obj.__func__ = None

        if func:
            obj.wrap(func)

        return obj


class Argument(_MethodDecorator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.group = None

    def wrap(self, wrapped):
        if isinstance(wrapped, Argument):
            group = Group()

            self.wrap(wrapped.__func__)
            group.wrap(wrapped.__func__)
            group.add(wrapped)
            group.add(self)

            # Don't duplicate help message
            wrapped.__doc__ = None
            self.__doc__ = None

            return group

        if isinstance(wrapped, Group):
            group = wrapped

            self.wrap(group.__func__)
            group.add(self)

            # Don't duplicate help message
            self.__doc__ = None

            return group

        return super().wrap(wrapped)

    def copy(self, *args, **kwargs):
        arg = super().copy(*args, **kwargs)
        arg.group = self.group

        return arg

    def __get__(self, instance, owner):
        arg = super().__get__(instance, owner)

        if isinstance(arg, inspect.BoundArguments):
            inputs = instance.__dict__.pop(self.__name__)
            func = super().__get__(instance, owner)
            instance.__dict__[self.__name__] = inputs

            if callable(func):
                try:
                    outputs = func(*inputs.args, **inputs.kwargs)
                except AttributeError as err:
                    raise RuntimeError(
                        "Argument method raised AttributeError"
                    ) from err

                if inspect.isgenerator(outputs):
                    try:
                        while True:
                            instance.__dict__[self.__name__] = next(outputs)
                    except StopIteration as err:
                        outputs = err.value

                instance.__dict__[self.__name__] = outputs
                return outputs

            return inputs

        return arg

    @property
    def __auto_kwargs(self):
        kwargs = {}

        if self.__func__ is not None:
            # Bind to anything to skip the "self" argument
            func = self.__func__.__get__(object(), object)

            act_kwargs, f_kwargs = filter_action_kwargs(
                self._kwargs,
                action=FunctionBindAction,
            )

            if f_kwargs:
                act_kwargs.setdefault("kwargs", {})
                act_kwargs["kwargs"].update(f_kwargs)
            f_args = act_kwargs.get("args", ())
            f_kwargs = act_kwargs.get("kwargs", {})

            partial = functools.partial(func, *f_args, **f_kwargs)
            sig = inspect.signature(partial, follow_wrapped=False)
            params = sig.parameters

            kwargs["action"] = FunctionBindAction
            kwargs["func"] = func

        else:
            params = {}

        nargs_min = 0
        nargs_max = 0
        var_args = None
        var_kwargs = None
        first_arg = next(iter(params.keys())).upper() if params else None

        for name, param in params.items():
            if param.kind == inspect.Parameter.VAR_POSITIONAL:
                var_args = param
                continue

            elif param.kind == inspect.Parameter.VAR_KEYWORD:
                var_kwargs = param
                continue

            elif param.kind == inspect.Parameter.KEYWORD_ONLY:
                # partial() objs accept kwargs that're already bound
                if name in f_kwargs:
                    continue

            if param.default == inspect.Parameter.empty:
                nargs_min += 1
                nargs_max += 1

            else:
                nargs_max += 1

        option_strings = self._args
        kwargs["dest"] = self.__name__

        doc = inspect.getdoc(self)
        if doc is not None:
            kwargs["help"] = doc.split("\n\n")[0]

        # attr = Argument()
        if self.__func__ is None and not option_strings:
            kwargs["nargs"] = 1

        # attr = Argument("--arg")
        elif self.__func__ is None:
            kwargs["nargs"] = "?"

        # func(a, *b)
        elif (var_args or var_kwargs) and nargs_min > 0:
            kwargs["nargs"] = "+"
            kwargs["metavar"] = (var_args or var_kwargs).name

        # func(*a)
        elif (var_args or var_kwargs) and nargs_min == 0:
            kwargs["nargs"] = "*"
            kwargs["metavar"] = (var_args or var_kwargs).name

        # func()
        elif (nargs_min, nargs_max) == (0, 0):
            kwargs["nargs"] = 0

        # func(a=None)
        elif (nargs_min, nargs_max) == (0, 1):
            kwargs["nargs"] = "?"
            kwargs["metavar"] = first_arg

        # func(a=None, b=None)
        elif nargs_min == 0:
            kwargs["nargs"] = "*"
            kwargs["metavar"] = first_arg

        # func(a, b=None)
        elif nargs_min != nargs_max:
            kwargs["nargs"] = "+"
            kwargs["metavar"] = first_arg

        # func(a, b)
        else:
            kwargs["nargs"] = nargs_min

            if option_strings:
                kwargs["metavar"] = tuple(params.keys())
            else:
                kwargs["metavar"] = first_arg

        def format_metavar(s):
            return s.replace("-","_").strip(" -_").upper()

        if "metavar" in kwargs:
            metavar = kwargs["metavar"]
            if isinstance(metavar, tuple):
                metavar = tuple(format_metavar(m) for m in metavar)
            else:
                metavar = format_metavar(metavar)
            kwargs["metavar"] = metavar

        return kwargs

    @property
    def __kwargs(self):
        kwargs = self.__auto_kwargs
        kwargs.update(self._kwargs)

        act_kwargs, f_kwargs = filter_action_kwargs(kwargs)

        act = act_kwargs.get("action", None)
        if isinstance(act, type) and issubclass(act, FunctionBindAction):
            if act_kwargs.get("count", False):
                act_kwargs["nargs"] = 0

            if f_kwargs:
                act_kwargs.setdefault("kwargs", {})
                act_kwargs["kwargs"].update(f_kwargs)

        return act_kwargs

    def build(self, parent):
        option_strings = self._args
        kwargs = self.__kwargs

        return parent.add_argument(*option_strings, **kwargs)

    def __property_from_kwargs(name):
        @property
        def prop(self):
            try:
                return self.__kwargs[name]
            except KeyError:
                raise AttributeError(
                    "Argument '{}' does not pass '{}' to add_argument"
                    .format(self.__name__, name)
                ) from None

        @prop.setter
        def prop(self, value):
            self._kwargs[name] = value

        @prop.deleter
        def prop(self):
            del self._kwargs[name]

        return prop

    @property
    def name_or_flags(self):
        return self._args or (self.__name__,)

    action = __property_from_kwargs("action")
    nargs = __property_from_kwargs("nargs")
    default = __property_from_kwargs("default")
    type = __property_from_kwargs("type")
    choices = __property_from_kwargs("choices")
    required = __property_from_kwargs("required")
    help = __property_from_kwargs("help")
    metavar = __property_from_kwargs("metavar")
    dest = __property_from_kwargs("dest")
    append = __property_from_kwargs("append")
    count = __property_from_kwargs("count")
    version = __property_from_kwargs("version")
    del __property_from_kwargs


class Group(_MethodDecorator):
    def __init__(self, *args, **kwargs):
        self._arguments = []
        self.group = None
        super().__init__(*args, **kwargs)

    def wrap(self, wrapped):
        if isinstance(wrapped, Argument):
            self.wrap(wrapped.__func__)
            self.add(wrapped)

            # Don't duplicate help message
            wrapped.__doc__ = None

            return self

        if isinstance(wrapped, Group):
            old_doc = self.__doc__
            self.wrap(wrapped.__func__)
            self.add(wrapped)

            # Don't override help message
            if old_doc is not None:
                wrapped.__doc__ = self.__doc__
                self.__doc__ = old_doc

            return self

        return super().wrap(wrapped)

    def copy(self, *args, **kwargs):
        grp = super().copy(*args, **kwargs)

        for arg in self._arguments:
            grp.add(arg)

        return grp

    def __get__(self, instance, owner):
        grp = super().__get__(instance, owner)

        if isinstance(grp, inspect.BoundArguments):
            inputs = instance.__dict__.pop(self.__name__)
            func = super().__get__(instance, owner)
            instance.__dict__[self.__name__] = inputs

            if callable(func):
                try:
                    outputs = func(*inputs.args, **inputs.kwargs)
                except AttributeError as err:
                    raise RuntimeError(
                        "Group method raised AttributeError"
                    ) from err

                if inspect.isgenerator(outputs):
                    try:
                        while True:
                            instance.__dict__[self.__name__] = next(outputs)
                    except StopIteration as err:
                        outputs = err.value

                instance.__dict__[self.__name__] = outputs
                return outputs

            return inputs

        return grp

    @property
    def __auto_kwargs(self):
        kwargs = {}

        doc = inspect.getdoc(self)
        if doc:
            blocks = doc.split("\n\n")
            kwargs["title"] = blocks[0].replace("\n", " ")
            if len(blocks) > 1:
                kwargs["description"] = "\n\n".join(blocks[1:])
        else:
            kwargs["title"] = self.__name__

        return kwargs

    @property
    def __kwargs(self):
        kwargs = self.__auto_kwargs
        kwargs.update(self._kwargs)

        return kwargs

    def build(self, parent):
        parser = parent.add_argument_group(*self._args, **self.__kwargs)

        items = list(self._arguments)
        while items:
            item = items.pop(0)
            item.__get__(self.__self__, type(self.__self__))

            # Argparse doesn't print help message for nested groups,
            # so we flatten them here.
            if isinstance(item, Group):
                items = list(item._arguments) + items
                continue

            item.build(parser)

        return parser

    def add(self, arg):
        self._arguments.append(arg)
        arg.group = self
        return arg

    def __property_from_kwargs(name):
        @property
        def prop(self):
            try:
                return self.__kwargs[name]
            except KeyError:
                raise AttributeError(
                    "Group '{}' does not pass '{}' to add_argument_group"
                    .format(self.__name__, name)
                ) from None

        @prop.setter
        def prop(self, value):
            self._kwargs[name] = value

        @prop.deleter
        def prop(self):
            del self._kwargs[name]

        return prop

    title = __property_from_kwargs("title")
    description = __property_from_kwargs("description")
    del __property_from_kwargs


class Subparsers(_MethodDecorator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._commands = []

    def wrap(self, func):
        if isinstance(func, CommandMeta):
            self.add(func)
            return self

        return super().wrap(func)

    @property
    def __auto_kwargs(self):
        kwargs = {}

        doc = inspect.getdoc(self)
        if doc:
            blocks = doc.split("\n\n")
            kwargs["title"] = blocks[0].replace("\n", " ")
            kwargs["description"] = "\n\n".join(blocks[1:])
        else:
            kwargs["title"] = self.__name__

        kwargs["dest"] = self.__name__

        return kwargs

    @property
    def __kwargs(self):
        kwargs = self.__auto_kwargs
        kwargs.update(self._kwargs)

        return kwargs

    def build(self, parent):
        subparsers = parent.add_subparsers(*self._args, **self.__kwargs)

        for cmd in self._commands:
            cmd.build(subparsers)

        return subparsers

    def add(self, cmd):
        self._commands.append(cmd)
        return cmd

    def __property_from_kwargs(name):
        @property
        def prop(self):
            try:
                return self.__kwargs[name]
            except KeyError:
                raise AttributeError(
                    "Subparsers '{}' does not pass '{}' to add_subparsers"
                    .format(self.__name__, name)
                ) from None

        @prop.setter
        def prop(self, value):
            self._kwargs[name] = value

        @prop.deleter
        def prop(self):
            del self._kwargs[name]

        return prop

    title = __property_from_kwargs("title")
    description = __property_from_kwargs("description")
    dest = __property_from_kwargs("dest")
    del __property_from_kwargs


def command_call(call):
    def __call__(self, **kwargs):
        for kwarg, value in kwargs.items():
            func = getattr(self, kwarg)
            if not callable(func):
                setattr(self, kwarg, value)
                continue

            if isinstance(value, inspect.BoundArguments):
                setattr(self, kwarg, value)
                continue

            sig = inspect.signature(func)
            var_args = None
            var_kwargs = None

            for name, param in sig.parameters.items():
                if param.kind == inspect.Parameter.VAR_POSITIONAL:
                    var_args = param

                elif param.kind == inspect.Parameter.VAR_KEYWORD:
                    var_kwargs = param
                    raise NotImplementedError

                elif param.kind == inspect.Parameter.KEYWORD_ONLY:
                    raise NotImplementedError

            # func(*a)
            # func(a, *b)
            if var_args and isinstance(value, (list, set, tuple)):
                value = sig.bind(*value)

            # func()
            elif len(sig.parameters) == 0:
                value = sig.bind() if value else None

            # func(a)
            # func(a=None)
            elif len(sig.parameters) == 1:
                value = sig.bind(value)

            # func(a, b)
            # func(a, b=None)
            # func(a=None, b=None)
            else:
                value = sig.bind(*value)

            setattr(self, kwarg, value)

        for cmd in reversed(type(self).__mro__):
            if not isinstance(cmd, CommandMeta):
                continue

            for arg_name, arg in cmd.arguments():
                arg.__self__ = None
                try:
                    func = arg.__func__.__get__(object(), object)
                    sig = inspect.signature(func)
                    self.__dict__.setdefault(arg_name, sig.bind())
                except:
                    self.__dict__.setdefault(arg_name, None)

            for grp_name, grp in cmd.groups():
                grp.__self__ = None
                for arg in grp._arguments:
                    arg.__self__ = None
                    if not hasattr(arg, "dest") or arg.dest == argparse.SUPPRESS:
                        continue
                    try:
                        func = arg.__func__.__get__(object(), object)
                        sig = inspect.signature(func)
                        self.__dict__.setdefault(arg.dest, sig.bind())
                    except:
                        self.__dict__.setdefault(arg.dest, None)

                if grp_name not in (
                    arg.dest for arg in grp._arguments
                    if hasattr(arg, "dest")
                ):
                    try:
                        # grp.__get__(self, type(self)) would mutate self
                        func = grp.__func__.__get__(object(), object)
                        sig = inspect.signature(func)
                        self.__dict__.setdefault(grp_name, sig.bind())
                    except:
                        self.__dict__.setdefault(grp_name, None)

        for cmd in reversed(type(self).__mro__):
            if not isinstance(cmd, CommandMeta):
                continue

            for arg_name, arg in cmd.arguments():
                getattr(self, arg_name, None)

            for grp_name, grp in cmd.groups():
                getattr(self, grp_name, None)

        return call(self)

    functools.update_wrapper(__call__, call)
    return __call__


class CommandExit(Exception):
    def __init__(
        self,
        message=None,
        output=None,
        returncode=1,
        errno=None,
    ):
        if errno is not None:
            if message:
                errmsg = "[Errno {}] {}".format(errno, message)
            else:
                errmsg = "[Errno {}]".format(errno)
        else:
            errmsg = message

        self.returncode = returncode
        self.output = output
        self.message = message
        self.errno = errno
        super().__init__(errmsg)

    def __repr__(self):
        return "{}(output={!r}, returncode={!r}, message={!r})".format(
            type(self).__qualname__,
            self.output,
            self.returncode,
            self.message,
        )


class CommandMeta(type):
    def __new__(mcls, name, bases, attrs, **kwargs):
        call = attrs.get("__call__", None)

        if call is not None:
            attrs["__call__"] = command_call(call)

        cls = super().__new__(mcls, name, bases, attrs)
        cls.__custom_kwargs = kwargs
        return cls

    @property
    def __call__(cls):
        cls_call = cls.__dict__.get("__call__", None)
        cls_call = getattr(cls_call, "__func__", cls_call)

        if inspect.isgeneratorfunction(cls_call):
            return cls.__generator_call
        else:
            return cls.__normal_call

    def __normal_call(cls, *args, **kwargs):
        instance = super().__call__()
        raise_exit = kwargs.pop("__raise_CommandExit", False)

        if hasattr(instance, "__enter__"):
            with instance as inst:
                retval = inst(*args, **kwargs)
        else:
            retval = instance(*args, **kwargs)

        if isinstance(retval, CommandExit):
            if raise_exit:
                raise retval
            elif retval.returncode != 0 and hasattr(instance, "_logger"):
                instance._logger.warning(retval.message)
                retval = retval.output
            else:
                retval = retval.output

        return retval

    def __generator_call(cls, *args, **kwargs):
        instance = super().__call__()
        raise_exit = kwargs.pop("__raise_CommandExit", False)

        if hasattr(instance, "__enter__"):
            with instance as inst:
                retval = yield from inst(*args, **kwargs)
        else:
            retval = yield from instance(*args, **kwargs)

        if isinstance(retval, CommandExit):
            if raise_exit:
                raise retval
            elif retval.returncode != 0 and hasattr(instance, "_logger"):
                instance._logger.warning(retval.message)
                retval = retval.output
            else:
                retval = retval.output

        return retval

    def main(cls, *argv):
        if len(argv) == 0:
            prog, *argv = sys.argv

        parser = cls.parser
        args = parser.parse_args(argv)
        command = getattr(args, "__command")
        kwargs = {
            k: v for k, v in vars(args).items()
            if k != "__command"
        }

        root_logger = logging.getLogger()
        root_logger.addHandler(logging.StreamHandler())
        root_logger.setLevel(logging.NOTSET)

        if hasattr(command, "_logger"):
            logger = command._logger
        else:
            logging.getLogger(cls.__module__)

        def log_error(err):
            is_debug_level = logger.isEnabledFor(logging.DEBUG)
            if not is_debug_level and err.__cause__ is not None:
                log_error(err.__cause__)
            logger.error(err, exc_info=is_debug_level)

        def print_out(output):
            if inspect.isgenerator(output):
                try:
                    while True:
                        print(next(output))
                except StopIteration as err:
                    output = err.value

            elif isinstance(output, list):
                print(*output, sep="\n")

            elif isinstance(output, tuple):
                print(*output, sep="\n")

            elif output is not None:
                print(output)

        try:
            output = command(__raise_CommandExit=True, **kwargs)
            print_out(output)

        except CommandExit as exit:
            if exit.returncode != 0:
                log_error(exit)
            else:
                logger.warning(exit)

            print_out(exit.output)
            sys.exit(exit.returncode)

        except Exception as err:
            log_error(err)
            sys.exit(1)

    def items(cls):
        def order(tup):
            attr, value = tup
            return (
                isinstance(value, Group),
                isinstance(value, Argument),
                isinstance(value, Subparsers),
                isinstance(value, CommandMeta),
            )
        pairs = (
            (k, v) for k, v in vars(cls).items()
            if not k.startswith("_")
        )
        return sorted(pairs, key=order, reverse=True)

    def groups(cls):
        yield from (
            (k, v) for k, v in cls.items()
            if isinstance(v, Group)
        )

    def arguments(cls):
        yield from (
            (k, v) for k, v in cls.items()
            if isinstance(v, Argument)
        )

    def subparsers(cls):
        yield from (
            (k, v) for k, v in cls.items()
            if isinstance(v, Subparsers)
        )

    def subcommands(cls):
        yield from (
            (k, v) for k, v in cls.items()
            if isinstance(v, CommandMeta)
        )

    def subcommand(cls, arg):
        if isinstance(arg, CommandMeta):
            name = arg.__name__
            while hasattr(cls, name):
                name = "{}_".format(name)
            setattr(cls, name, arg)
            return arg

        def add_subcommand(cmd):
            name = arg.replace("-", "_")
            while hasattr(cls, name):
                name = "{}_".format(name)
            setattr(cls, name, cmd)
            cmd.__name__ = arg
            return cmd

        return add_subcommand

    @property
    def __auto_kwargs(cls):
        kwargs = {}

        doc = inspect.getdoc(cls)
        if doc is not None:
            blocks = doc.split("\n\n")

            for i, block in enumerate(blocks):
                if block.strip("- ") == "":
                    kwargs["help"] = blocks[0]
                    kwargs["description"] = "\n\n".join(blocks[:i])
                    kwargs["epilog"] = "\n\n".join(blocks[i+1:])
                    break
            else:
                kwargs["help"] = blocks[0]
                kwargs["description"] = doc

        kwargs["argument_default"] = argparse.SUPPRESS
        kwargs["formatter_class"] = argparse.RawDescriptionHelpFormatter
        kwargs["conflict_handler"] = "resolve"

        return kwargs

    @property
    def __kwargs(cls):
        kwargs = cls.__auto_kwargs
        kwargs.update(cls.__custom_kwargs)

        return kwargs

    @property
    def parser(cls):
        return cls.__build()

    def __build(cls, parent=None):
        kwargs = cls.__kwargs

        if parent is None:
            kwargs.pop("help", None)
            parser = argparse.ArgumentParser(**kwargs)

        else:
            parser = parent.add_parser(cls.__name__, **kwargs)

        subparsers = None

        for attr, value in cls.items():
            if isinstance(value, Group):
                group = getattr(cls, attr)
                if group.group is None:
                    group.build(parser)

            elif isinstance(value, Argument):
                arg = getattr(cls, attr)
                if arg.group is None:
                    arg.build(parser)

            elif isinstance(value, Subparsers):
                obj = getattr(cls, attr)
                subparsers = obj.build(parser)

            elif isinstance(value, CommandMeta):
                cmd = getattr(cls, attr)
                cmd.__self__ = cls

                if subparsers is None:
                    subparsers = parser.add_subparsers()
                cmd.__build(subparsers)

        parser.set_defaults(__command=cls)

        return parser

    build = __build

    def __property_from_kwargs(name):
        @property
        def prop(cls):
            if name in cls.__dict__:
                return cls.__dict__[name]

            try:
                return cls.__kwargs[name]
            except KeyError:
                raise AttributeError(
                    "Command '{}' does not pass '{}' to ArgumentParser"
                    .format(cls.__name__, name)
                ) from None

        @prop.setter
        def prop(cls, value):
            if name in cls.__dict__:
                prop = getattr(type(cls), name)
                delattr(type(cls), name)
                super().__setattr__(name, value)
                setattr(type(cls), name, prop)
            else:
                cls.__custom_kwargs[name] = value

        @prop.deleter
        def prop(cls):
            if name in cls.__dict__:
                prop = getattr(type(cls), name)
                delattr(type(cls), name)
                super().__delattr__(name)
                setattr(type(cls), name, prop)
            else:
                del cls.__custom_kwargs[name]

        return prop

    prog = __property_from_kwargs("prog")
    usage = __property_from_kwargs("usage")
    description = __property_from_kwargs("description")
    epilog = __property_from_kwargs("epilog")
    parents = __property_from_kwargs("parents")
    formatter_class = __property_from_kwargs("formatter_class")
    prefix_chars = __property_from_kwargs("prefix_chars")
    fromfile_prefix_chars = __property_from_kwargs("fromfile_prefix_chars")
    argument_default = __property_from_kwargs("argument_default")
    conflict_handler = __property_from_kwargs("conflict_handler")
    add_help = __property_from_kwargs("add_help")
    allow_abbrev = __property_from_kwargs("allow_abbrev")
    help = __property_from_kwargs("help")
    del __property_from_kwargs


class Command(metaclass=CommandMeta):
    def __init__(self):
        self.exitstack = contextlib.ExitStack()

    def __enter__(self):
        self.exitstack.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return self.exitstack.__exit__(exc_type, exc_value, traceback)



#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools depthchargectl remove subcommand
# Copyright (C) 2020-2022 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

import argparse
import logging
import subprocess

from pathlib import Path

from depthcharge_tools import (
    __version__,
)
from depthcharge_tools.utils.argparse import (
    Command,
    Argument,
    Group,
    CommandExit,
)

from depthcharge_tools.depthchargectl import depthchargectl


class BootedPartitionError(CommandExit):
    def __init__(self, partition):
        self.partition = partition
        super().__init__(
            "Refusing to disable currently booted partition '{}'."
            .format(partition)
        )


@depthchargectl.subcommand("remove")
class depthchargectl_remove(
    depthchargectl,
    prog="depthchargectl remove",
    usage="%(prog)s [options] (KERNEL_VERSION | IMAGE)",
    add_help=False,
):
    """Remove images and disable partitions containing them."""

    _logger = depthchargectl._logger.getChild("remove")
    config_section = "depthchargectl/remove"

    @depthchargectl.board.copy()
    def board(self, codename=""):
        """Assume we're running on the specified board"""
        # We can disable partitions without knowing the board.
        try:
            return super().board
        except Exception as err:
            self.logger.warning(err)
            return None

    @Group
    def positionals(self):
        """Positional arguments"""

        if self.image is not None and self.kernel_version is not None:
            raise ValueError(
                "Image and kernel_version arguments are mutually exclusive."
            )

        if self.image is not None:
            image = self.image
        else:
            image = self.kernel_version

        if isinstance(image, str):
            # This can be run after the kernel is uninstalled, where the
            # version would no longer be valid, so don't check for that.
            # Instead just check if we have it as an image.
            img = (self.images_dir / "{}.img".format(image)).resolve()
            if img.parent == self.images_dir and img.is_file():
                self.logger.info(
                    "Disabling partitions for kernel version '{}'."
                    .format(image)
                )
                self.image = img
                self.kernel_version = image

            else:
                self.image = Path(image).resolve()
                self.kernel_version = None
                self.logger.info(
                    "Disabling partitions for depthcharge image '{}'."
                    .format(image)
                )

        if not self.image.is_file():
            raise TypeError(
                "Image to remove '{}' is not a file."
                .format(self.image)
            )

    @positionals.add
    @Argument(dest=argparse.SUPPRESS, nargs=0)
    def kernel_version(self, kernel_version):
        """Installed kernel version to disable."""
        return kernel_version

    @positionals.add
    @Argument
    def image(self, image):
        """Depthcharge image to disable."""
        return image

    @Group
    def options(self):
        """Options"""

    @options.add
    @Argument("-f", "--force", force=True)
    def force(self, force=False):
        """Allow disabling the currently booted partition."""
        return force

    def __call__(self):
        image = self.image

        # When called with --vblockonly vbutil_kernel creates a file of
        # size 64KiB == 0x10000.
        image_vblock = image.read_bytes()[:0x10000]

        partitions = depthchargectl.list(
            root=self.root,
            root_mountpoint=self.root_mountpoint,
            boot_mountpoint=self.boot_mountpoint,
            config=self.config,
            board=self.board,
            tmpdir=self.tmpdir / "list",
            images_dir=self.images_dir,
            vboot_keyblock=self.vboot_keyblock,
            vboot_public_key=self.vboot_public_key,
            vboot_private_key=self.vboot_private_key,
            kernel_cmdline=self.kernel_cmdline,
            ignore_initramfs=self.ignore_initramfs,
            verbosity=self.verbosity,
        )

        self.logger.info(
            "Searching for Chrome OS Kernel partitions containing '{}'."
            .format(image)
        )
        badparts = []
        error_disks = []

        for part in partitions:
            self.logger.info("Checking partition '{}'.".format(part))

            # It's OK to check only the vblock header, as that
            # contains signatures on the content and those will be
            # different if the content is different.
            with part.path.open("rb") as p:
                if p.read(0x10000) == image_vblock:
                    try:
                        if part.attribute:
                            badparts.append(part)
                    except subprocess.CalledProcessError as err:
                        self.logger.warning(
                            "Couldn't get attribute for partition '{}'."
                            .format(part)
                        )
                        self.logger.debug(
                            err,
                            exc_info=self.logger.isEnabledFor(
                                logging.DEBUG,
                            ),
                        )

        current = self.diskinfo.by_kern_guid()
        if current in badparts:
            if self.force:
                self.logger.warning(
                    "Deactivating the currently booted partition '{}'. "
                    "This might make your system unbootable."
                    .format(current)
                )
            else:
                raise BootedPartitionError(current)

        done_parts = []
        error_parts = []
        for part in badparts:
            self.logger.info("Deactivating '{}'.".format(part))
            try:
                depthchargectl.bless(
                    partition=part,
                    bad=True,
                    root=self.root,
                    root_mountpoint=self.root_mountpoint,
                    boot_mountpoint=self.boot_mountpoint,
                    config=self.config,
                    board=self.board,
                    tmpdir=self.tmpdir / "bless",
                    images_dir=self.images_dir,
                    vboot_keyblock=self.vboot_keyblock,
                    vboot_public_key=self.vboot_public_key,
                    vboot_private_key=self.vboot_private_key,
                    kernel_cmdline=self.kernel_cmdline,
                    ignore_initramfs=self.ignore_initramfs,
                    verbosity=self.verbosity,
                )
            except Exception as err:
                error_parts.append(part)
                self.logger.debug(
                    err,
                    exc_info=self.logger.isEnabledFor(logging.DEBUG),
                )
                continue

            done_parts.append(part)
            self.logger.warning("Deactivated '{}'.".format(part))

        if image.parent == self.images_dir and not error_disks and not error_parts:
            self.logger.info(
                "Image '{}' is in images dir, deleting."
                .format(image)
            )
            image.unlink()
            self.logger.warning("Deleted image '{}'.".format(image))

        else:
            self.logger.info(
                "Not deleting image file '{}'."
                .format(image)
            )

        output = badparts or None

        error_msg = []
        if error_disks:
            error_msg.append(
                "Couldn't disable partitions for disks {}."
                .format(", ".join(str(d) for d in error_disks))
            )

        if error_parts:
            error_msg.append(
                "Couldn't disable partitions {}."
                .format(", ".join(str(d) for d in error_parts))
            )

        if error_msg:
            return CommandExit(
                message="\n".join(error_msg),
                output=done_parts,
                returncode=1,
            )

        if not output:
            self.logger.warning(
                "No active partitions contain the given image."
            )

        return output

    global_options = depthchargectl.global_options
    config_options = depthchargectl.config_options



#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools depthchargectl target subcommand
# Copyright (C) 2020-2022 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

import argparse
import logging
import subprocess
import sys
import types

from depthcharge_tools import __version__
from depthcharge_tools.utils.argparse import (
    Command,
    Argument,
    Group,
    CommandExit,
)
from depthcharge_tools.utils.os import (
    Disk,
    CrosPartition,
    Partition,
)
from depthcharge_tools.utils.string import (
    parse_bytesize,
)

from depthcharge_tools.depthchargectl import depthchargectl


class NotABlockDeviceError(CommandExit):
    def __init__(self, device):
        message = (
            "Target '{}' is not a valid block device."
            .format(device)
        )

        self.device = device
        super().__init__(message=message, returncode=2)


class NotCrosPartitionError(CommandExit):
    def __init__(self, partition):
        message = (
            "Partition '{}' is not of type Chrome OS Kernel."
            .format(partition)
        )

        self.partition = partition
        super().__init__(message=message, returncode=5)


class BootedPartitionError(CommandExit):
    def __init__(self, partition):
        message = (
            "Partition '{}' is the currently booted parttiion."
            .format(partition)
        )

        self.partition = partition
        super().__init__(message=message, returncode=6)


class PartitionSizeTooSmallError(CommandExit):
    def __init__(self, partition, part_size, min_size):
        message = (
            "Partition '{}' ('{}' bytes) is smaller than '{}' bytes."
            .format(partition, part_size, min_size)
        )

        self.partition = partition
        self.part_size = part_size
        self.min_size = min_size
        super().__init__(message=message, returncode=7)


class NoUsableCrosPartition(CommandExit):
    def __init__(self):
        message = (
            "No usable Chrome OS Kernel partition found "
            "for given input arguments."
        )

        super().__init__(message=message, output=None)


@depthchargectl.subcommand("target")
class depthchargectl_target(
    depthchargectl,
    prog="depthchargectl target",
    usage="%(prog)s [options] [PARTITION | DISK ...]",
    add_help=False,
):
    """Choose or validate a ChromeOS Kernel partition to use."""

    _logger = depthchargectl._logger.getChild("target")
    config_section = "depthchargectl/target"

    @depthchargectl.board.copy()
    def board(self, codename=""):
        """Assume we're running on the specified board"""
        # We can target partitions without knowing the board.
        try:
            return super().board
        except Exception as err:
            self.logger.warning(err)
            return None

    @Group
    def positionals(self):
        """Positional arguments"""

        disks = list(self.disks)
        partitions = list(self.partitions)

        # The inputs can be a mixed list of partitions and disks,
        # separate the two.
        for d in list(disks):
            try:
                partitions.append(Partition(d))
                self.logger.info("Using target '{}' as a partition.".format(d))
                disks.remove(d)
            except:
                pass

        self.disks = disks
        self.partitions = partitions

    @positionals.add
    @Argument(metavar="PARTITION", nargs=0)
    def partitions(self, *partitions):
        """Chrome OS kernel partition to validate."""
        return partitions

    @positionals.add
    @Argument(metavar="DISK")
    def disks(self, *disks):
        """Disks to search for an appropriate Chrome OS kernel partition."""
        return disks

    @Group
    def options(self):
        """Options"""

    @options.add
    @Argument("-s", "--min-size", nargs=1)
    def min_size(self, bytes_=None):
        """Target partitions larger than this size."""
        if bytes_ is None:
            return 0x10000

        return parse_bytesize(bytes_)

    @options.add
    @Argument("--allow-current", allow=True)
    def allow_current(self, allow=False):
        """Allow targeting the currently booted partition."""
        return allow

    @options.add
    @Argument("-a", "--all-disks", all_disks=True)
    def all_disks(self, all_disks=False):
        """Target partitions on all disks."""
        return all_disks

    def __call__(self):
        disks = list(self.disks)
        partitions = list(self.partitions)

        # We will need to check partitions against this if allow_current
        # is false.
        current = self.diskinfo.by_kern_guid()

        # Given a single partition, check if the partition is valid.
        if len(partitions) == 1 and len(disks) == 0:
            part = partitions[0]

            self.logger.info("Checking if target partition is writable.")
            if part.path is not None and not part.path.is_block_device():
                raise NotABlockDeviceError(part.path)

            self.logger.info("Checking if targeted partition's disk is writable.")
            if not part.disk.path.is_block_device():
                raise NotABlockDeviceError(part.disk.path)

            self.logger.info(
                "Checking if targeted partition's type is Chrome OS Kernel."
            )
            if part not in part.disk.cros_partitions():
                raise NotCrosPartitionError(part)

            self.logger.info(
                "Checking if targeted partition is currently booted one."
            )
            if current is not None and not self.allow_current:
                if part.path == current.path:
                    raise BootedPartitionError(part)

            self.logger.info(
                "Checking if targeted partition is bigger than given "
                "minimum size."
            )
            if self.min_size is not None and part.size < self.min_size:
                raise PartitionSizeTooSmallError(
                    part,
                    part.size,
                    self.min_size,
                )

        # For arguments which are disks, search all their partitions.
        # If no disks or partitions were given, search bootable disks.
        # Search all disks if explicitly asked.
        if disks or not partitions or self.all_disks:
            partitions += depthchargectl.list(
                disks=disks,
                all_disks=self.all_disks,
                root=self.root,
                root_mountpoint=self.root_mountpoint,
                boot_mountpoint=self.boot_mountpoint,
                config=self.config,
                board=self.board,
                tmpdir=self.tmpdir / "list",
                images_dir=self.images_dir,
                vboot_keyblock=self.vboot_keyblock,
                vboot_public_key=self.vboot_public_key,
                vboot_private_key=self.vboot_private_key,
                kernel_cmdline=self.kernel_cmdline,
                ignore_initramfs=self.ignore_initramfs,
                verbosity=self.verbosity,
            )

        good_partitions = []
        for p in partitions:
            if self.min_size is not None and p.size < self.min_size:
                self.logger.warning(
                    "Skipping partition '{}' as too small."
                    .format(p)
                )
                continue

            if current is not None and not self.allow_current:
                if p.path == current.path:
                    self.logger.info(
                        "Skipping currently booted partition '{}'."
                        .format(p)
                    )
                    continue

            self.logger.info("Partition '{}' is usable.".format(p))
            good_partitions.append(
                CrosPartition(p.disk.path, partno=p.partno),
            )

        # Get the least-successful, least-priority, least-tries-left
        # partition in that order of preference.
        if good_partitions:
            return min(good_partitions)
        else:
            return NoUsableCrosPartition()

    global_options = depthchargectl.global_options
    config_options = depthchargectl.config_options


